# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remote corpus manager."""

from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.google_cloud_utils import gsutil

from . import remote_process_host


class RemoteGSUtilRunner(gsutil.GSUtilRunner):
  """Remote GSUtil runner."""

  def __init__(self):
    super(
        RemoteGSUtilRunner,
        self).__init__(_process_runner=remote_process_host.RemoteProcessRunner)


class RemoteFuzzTargetCorpus(corpus_manager.FuzzTargetCorpus):
  """libFuzzer corpus sync that runs on untrusted bot."""

  def __init__(self, fuzzer_name, fuzzer_executable_name, quarantine=False):
    super(RemoteFuzzTargetCorpus, self).__init__(
        fuzzer_name,
        fuzzer_executable_name,
        quarantine,
        # Never log results for remote corpora since the state is on the worker.
        log_results=False,
        _gsutil_runner=RemoteGSUtilRunner)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The initialization script for Windows. It is run before running a task."""

import os
import subprocess

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.init_scripts import init_runner
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

DEFAULT_FAIL_RETRIES = 5
DEFAULT_FAIL_WAIT = 5

TEMP_DIRECTORIES = [
    r'%TEMP%', r'%USERPROFILE%\AppVerifierLogs', r'%USERPROFILE%\Downloads',
    r'%WINDIR%\Temp',
    r'C:\Program Files (x86)\Windows Kits\10\Debuggers\x86\sym',
    r'C:\Program Files (x86)\Windows Kits\10\Debuggers\x64\sym'
]


def clean_temp_directories():
  """Clean temporary directories."""
  for temp_directory in TEMP_DIRECTORIES:
    temp_directory_full_path = os.path.abspath(
        os.path.expandvars(temp_directory))
    shell.remove_directory(
        temp_directory_full_path, recreate=True, ignore_errors=True)


@retry.wrap(
    retries=DEFAULT_FAIL_RETRIES,
    delay=DEFAULT_FAIL_WAIT,
    function='init_scripts.windows.remount_if_needed')
def remount_if_needed():
  """Remount nfs volume if it is not working."""
  nfs_root = environment.get_value('NFS_ROOT')
  if not nfs_root:
    return

  nfs_host = environment.get_value('NFS_HOST')
  nfs_volume = environment.get_value('NFS_VOLUME')

  check_file_path = os.path.join(nfs_root, 'check')
  if os.path.exists(check_file_path):
    # Volume is mounted correctly and readable, bail out.
    return

  # Un-mount the nfs drive first. Ignore the return code as we might have
  # not mounted the drive at all.
  subprocess.call(['umount', '-f', nfs_root])

  # Mount the nfs drive.
  logs.log_warn('Trying to remount the NFS volume.')

  nfs_volume_path = '%s:/%s' % (nfs_host, nfs_volume)
  subprocess.check_call([
      'mount', '-o', 'anon', '-o', 'nolock', '-o', 'retry=10', nfs_volume_path,
      nfs_root
  ])

  if os.path.exists(check_file_path):
    # Volume is mounted correctly and readable, bail out.
    return

  # Update check file if needed.
  utils.write_data_to_file('ok', check_file_path)

  # Make sure that check file exists.
  if not os.path.exists(check_file_path):
    raise Exception('Failed to write check file on nfs volume.')


def run():
  """Run the initialization for Windows."""
  init_runner.run()
  clean_temp_directories()
  remount_if_needed()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the coverage_uploader module."""

import datetime
import os

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.fuzzing import coverage_uploader
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


def _mock_config_get(_, param):
  """Handle test configuration options."""
  if param == 'coverage.fuzzer-testcases.bucket':
    return 'test-coverage-testcases'

  return None


class FakeGSUtilRunner(object):
  """Fake gsutil runner for testing."""
  rsync_calls = []

  def rsync(self, source, destination, exclusion_pattern=None):
    FakeGSUtilRunner.rsync_calls.append((source, destination,
                                         exclusion_pattern))


class UploadTestsToCloudStorageTest(fake_filesystem_unittest.TestCase):
  """Tests for upload_tests_to_cloud_storage."""

  def setUp(self):
    """Setup for upload tests to cloud storage test."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'clusterfuzz._internal.config.local_config.ProjectConfig.get',
        'clusterfuzz._internal.datastore.locks.acquire_lock',
        'clusterfuzz._internal.datastore.locks.release_lock',
        'clusterfuzz._internal.google_cloud_utils.gsutil.GSUtilRunner',
        'clusterfuzz._internal.google_cloud_utils.storage.list_blobs',
        'clusterfuzz._internal.google_cloud_utils.storage.read_data',
        'clusterfuzz._internal.google_cloud_utils.storage.write_data',
    ])

    test_utils.set_up_pyfakefs(self)

    self.mock.write_data.return_value = True
    self.mock.utcnow.side_effect = lambda: datetime.datetime(2018, 11, 1, 0, 0)

    FakeGSUtilRunner.rsync_calls = []
    self.mock.GSUtilRunner.side_effect = FakeGSUtilRunner
    self.mock.get.side_effect = _mock_config_get

    os.environ['BOT_NAME'] = 'test-bot'
    os.environ['BOT_TMPDIR'] = '/tmp'
    os.environ['FAIL_RETRIES'] = '1'

  def test_tests_created_in_correct_bucket(self):
    """Ensure that we invoke gsutil correctly to store tests."""
    files = [
        '/testcases/a/file1.txt', '/testcases/file2.txt',
        '/something/b/file3.txt', '/data/f/g/file4.txt'
    ]
    coverage_uploader.upload_testcases_if_needed('test_fuzzer', files,
                                                 '/testcases', '/data')

    self.mock.write_data.assert_called_with(
        b'a/file1.txt\nfile2.txt\nf/g/file4.txt',
        'gs://test-coverage-testcases/2018-11-01/test_fuzzer/'
        '5b680a295e1f3a81160a0bd71ca2abbcb8d19521/file_list.txt')
    self.assertEqual(
        FakeGSUtilRunner.rsync_calls,
        [('/testcases', 'gs://test-coverage-testcases/'
          '2018-11-01/test_fuzzer/5b680a295e1f3a81160a0bd71ca2abbcb8d19521',
          None),
         ('/data', 'gs://test-coverage-testcases/'
          '2018-11-01/test_fuzzer/5b680a295e1f3a81160a0bd71ca2abbcb8d19521',
          '(?!.*fuzz-)'),
         ('/data', 'gs://test-coverage-testcases/'
          '2018-11-01/test_fuzzer/5b680a295e1f3a81160a0bd71ca2abbcb8d19521',
          '(?!.*resource)')])

  def test_empty_testcases_list(self):
    """Ensure that we do nothing when we have no testcases."""
    coverage_uploader.upload_testcases_if_needed('test_fuzzer', [],
                                                 '/testcases', '/data')
    self.assertEqual(self.mock.write_data.call_count, 0)
    self.assertEqual(FakeGSUtilRunner.rsync_calls, [])

  def test_large_testcase_list(self):
    """Ensure that we cap number of uploaded testcases."""
    files = ['/testcases/file%s' % i for i in range(20000)]
    coverage_uploader.upload_testcases_if_needed('test_fuzzer', files,
                                                 '/testcases', '/data')

    filtered_files_list = '\n'.join(
        ['file%s' % i for i in range(1000)]).encode('utf-8')
    self.mock.write_data.assert_called_with(
        filtered_files_list,
        'gs://test-coverage-testcases/2018-11-01/test_fuzzer/'
        '5b680a295e1f3a81160a0bd71ca2abbcb8d19521/file_list.txt')
    self.assertEqual(
        FakeGSUtilRunner.rsync_calls,
        [('/testcases', 'gs://test-coverage-testcases/'
          '2018-11-01/test_fuzzer/5b680a295e1f3a81160a0bd71ca2abbcb8d19521',
          None)])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Android emulator installation and management."""

import os
import time

from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.platforms.android import device
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from local.butler.reproduce_tool import errors
from local.butler.reproduce_tool import prompts

ADB_DEVICES_SEPARATOR_STRING = 'List of devices attached'
EMULATOR_RELATIVE_PATH = os.path.join('local', 'bin', 'android-sdk', 'emulator',
                                      'emulator')


def start_emulator():
  """Return a ProcessRunner configured to start the Android emulator."""
  root_dir = environment.get_value('ROOT_DIR')

  runner = new_process.ProcessRunner(
      os.path.join(root_dir, EMULATOR_RELATIVE_PATH),
      ['-avd', 'TestImage', '-writable-system', '-partition-size', '2048'])
  emulator_process = runner.run()

  # If we run adb commands too soon after the emulator starts, we may see
  # flake or errors. Delay a short while to account for this.
  # TODO(mbarbella): This is slow and flaky, but wait-for-device isn't usable if
  # another device is connected (as we don't know the serial yet). Find a better
  # solution.
  time.sleep(30)

  return emulator_process


def get_devices():
  """Get a list of all connected Android devices."""
  adb_runner = new_process.ProcessRunner(adb.get_adb_path())
  result = adb_runner.run_and_wait(additional_args=['devices'])

  if result.return_code:
    raise errors.ReproduceToolUnrecoverableError('Unable to run adb.')

  # Ignore non-device lines (those before "List of devices attached").
  store_devices = False
  devices = []
  for line in result.output.splitlines():
    if line == ADB_DEVICES_SEPARATOR_STRING:
      store_devices = True
      continue
    if not store_devices or not line:
      continue

    devices.append(line.split()[0])

  return devices


def prepare_environment(disable_android_setup):
  """Additional environment overrides needed to run on an Android device."""
  environment.set_value('OS_OVERRIDE', 'ANDROID')

  # Bail out if we can't determine which Android device to use.
  serial = environment.get_value('ANDROID_SERIAL')
  if not serial:
    devices = get_devices()
    if len(devices) == 1:
      serial = devices[0]
      environment.set_value('ANDROID_SERIAL', serial)
    elif not devices:
      raise errors.ReproduceToolUnrecoverableError(
          'No connected Android devices were detected. Run with the -e '
          'argument to use an emulator.')
    else:
      raise errors.ReproduceToolUnrecoverableError(
          'You have multiple Android devices or emulators connected. Please '
          'set the ANDROID_SERIAL environment variable and try again.\n\n'
          'Attached devices: ' + ', '.join(devices))

  print('Warning: this tool will make changes to settings on the connected '
        'Android device with serial {serial} that could result in data '
        'loss.'.format(serial=serial))
  willing_to_continue = prompts.get_boolean(
      'Are you sure you want to continue?')
  if not willing_to_continue:
    raise errors.ReproduceToolUnrecoverableError(
        'Bailing out to avoid changing settings on the connected device.')

  # Push the test case and build APK to the device.
  apk_path = environment.get_value('APP_PATH')
  device.update_build(
      apk_path, should_initialize_device=not disable_android_setup)

  device.push_testcases_to_device()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for AFL launcher.py."""

import os
import re
import shutil
import unittest

import mock

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.afl import fuzzer
from clusterfuzz._internal.bot.fuzzers.afl import launcher
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.tests.core.bot.fuzzers.afl.afl_launcher_test import \
    dont_use_strategies
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
TEMP_DIRECTORY = os.path.join(TEST_PATH, 'temp')
DATA_DIRECTORY = os.path.join(TEST_PATH, 'data')


def clear_temp_dir():
  """Clear temp directory."""
  if os.path.exists(TEMP_DIRECTORY):
    shutil.rmtree(TEMP_DIRECTORY)


def create_temp_dir():
  """Create temp directory."""
  os.mkdir(TEMP_DIRECTORY)


def get_fuzz_timeout(fuzz_time):
  """Return timeout for fuzzing."""
  return (fuzz_time + launcher.AflRunner.AFL_CLEAN_EXIT_TIME +
          launcher.AflRunner.SIGTERM_WAIT_TIME)


def no_errors(f):
  """Decorator that asserts neither metrics.logs.log_error nor
  metrics.logs.log_fatal_and_exit were called."""

  def call_f(self, *args, **kwargs):
    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])

    result = f(self, *args, **kwargs)
    self.assertEqual(0, self.mock.log_error.call_count)
    return result

  return call_f


def setup_testcase_and_corpus(testcase, corpus, fuzz=False):
  """Setup testcase and corpus."""
  copied_testcase_path = os.path.join(TEMP_DIRECTORY, testcase)
  shutil.copy(os.path.join(DATA_DIRECTORY, testcase), copied_testcase_path)

  copied_corpus_path = os.path.join(TEMP_DIRECTORY, corpus)
  src_corpus_path = os.path.join(DATA_DIRECTORY, corpus)

  if os.path.exists(src_corpus_path):
    shutil.copytree(src_corpus_path, copied_corpus_path)
  else:
    os.mkdir(copied_corpus_path)
    with open(os.path.join(copied_corpus_path, fuzzer.AFL_DUMMY_INPUT),
              'w') as f:
      f.write(' ')

  if fuzz:
    os.environ['FUZZ_CORPUS_DIR'] = copied_corpus_path

  return copied_testcase_path


def run_launcher(*args):
  """Run launcher.py."""
  mock_stdout = test_utils.MockStdout()

  os.environ['FUZZ_TARGET'] = args[1]
  with mock.patch('sys.stdout', mock_stdout):
    launcher.main(['launcher.py'] + list(args))

  return mock_stdout.getvalue()


def mocked_is_testcase(path):
  """Mocked version of AflFuzzOutputDirectory.is_testcase that looks for "COLON"
  instead of ":" because this repo cannot be used on windows if it has filenames
  with ":"."""
  testcase_regex = re.compile(r'idCOLON\d{6},.+')
  return (os.path.isfile(path) and
          bool(re.match(testcase_regex, os.path.basename(path))))


def mocked_fuzz(runner):
  """Mocked version of AflRunner.fuzz."""
  fuzz_args = runner.generate_afl_args()

  runner._fuzz_args = fuzz_args  # pylint: disable=protected-access
  engine_common.recreate_directory(runner.afl_output.output_directory)
  runner._fuzzer_stderr = ''  # pylint: disable=protected-access

  # Create the queue directory within AFL's output directory.
  queue = runner.afl_output.queue
  engine_common.recreate_directory(queue)
  new_corpus_dir = os.path.join(DATA_DIRECTORY, 'merge_new_corpus')
  for filename in os.listdir(new_corpus_dir):
    src = os.path.join(new_corpus_dir, filename)
    dst = os.path.join(queue, filename)
    shutil.copy(src, dst)

  return new_process.ProcessResult(
      command=[], return_code=0, output='', time_executed=1)


@unittest.skipIf(not environment.get_value('AFL_INTEGRATION_TESTS'),
                 'AFL_INTEGRATION_TESTS=1 must be set')
class BaseLauncherTest(unittest.TestCase):
  """Base AFL launcher tests."""

  def setUp(self):
    os.environ['BUILD_DIR'] = DATA_DIRECTORY
    os.environ['FUZZ_INPUTS_DISK'] = TEMP_DIRECTORY
    os.environ['FAIL_RETRIES'] = '1'

    test_helpers.patch_environ(self)
    dont_use_strategies(self)

    # Make it easy to assert if things were logged.
    test_helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log', 'os.getpid',
        'clusterfuzz._internal.bot.fuzzers.afl.launcher.rand_cmplog_level',
        'clusterfuzz._internal.bot.fuzzers.afl.launcher.rand_schedule'
    ])
    self.mock.rand_cmplog_level.return_value = '2'
    self.mock.rand_schedule.return_value = 'fast'
    self.logged_messages = []

    def mocked_log(message, **kwargs):  # pylint: disable=unused-argument
      self.logged_messages.append(message)

    self.mock.log.side_effect = mocked_log
    self.mock.getpid.return_value = 1337

    clear_temp_dir()
    create_temp_dir()

  def tearDown(self):
    clear_temp_dir()

  def _test_abnormal_return_code(self):
    """Test that abnormal return codes from single runs of the fuzz target (eg:
    not 0 or 1, which is ASAN's return code for errors) are logged."""
    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])
    testcase_path = setup_testcase_and_corpus('crash', 'empty_corpus')
    run_launcher(testcase_path, 'return_code_255')
    self.mock.log_error.assert_called_with(
        'AFL target exited with abnormal exit code: 255.',
        output='ERROR: returning 255\n')

  def _test_libfuzzerize_corpus(self, mock_get_timeout):
    """Test that libfuzzerize_corpus properly merges new testcases back into
    the corpus."""
    mock_get_timeout.return_value = get_fuzz_timeout(5.0)
    testcase_path = setup_testcase_and_corpus(
        'empty', 'input_corpus', fuzz=True)
    input_corpus = os.environ['FUZZ_CORPUS_DIR']

    corpus_path = os.path.join(DATA_DIRECTORY, 'merge_initial_corpus')
    for filename in os.listdir(corpus_path):
      src = os.path.join(corpus_path, filename)
      dst = os.path.join(input_corpus, filename)
      shutil.copy(src, dst)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.afl.launcher.AflRunnerCommon.fuzz',
        'clusterfuzz._internal.bot.fuzzers.afl.launcher.AflFuzzOutputDirectory.is_testcase'
    ])

    self.mock.fuzz.side_effect = mocked_fuzz
    self.mock.is_testcase.side_effect = mocked_is_testcase

    run_launcher(testcase_path, 'test_fuzzer', input_corpus)
    self.assertEqual(
        sorted([
            # Ensure that smaller files are favored. This file has the same
            # coverage as
            # 'idCOLON000000,larger_same_cov_idCOLON000001,non_initial_cov' but
            # is smaller, so merge should use 'idCOLON000001,non_initial_cov'
            # and not
            # 'idCOLON000000,larger_same_cov_idCOLON000001,non_initial_cov'.
            # This filename is the hash of idCOLON000001,non_initial_cov.
            '3dce8306f3c1810d5d81ed5ebb0ccea947277a61',
            # This file reaches unique paths and should therefore be merged.
            # This filename is the hash of idCOLON000002,unique_cov.
            '4c1fe29b1a967d34a6fcc078b8fb653dd807dee7',
            # Ensure we aren't removing anything from the input corpus. Thus
            # even though 'in1' and 'nearly_empty' are redundant they should
            # both be here Similarly, even though 'in1' and 'nearly_empty' have
            # the same coverage as idCOLON000003,nearly_empty_copy (from
            # merge_new_corpus/) and '254_ascending_and_0xa' has the same
            # coverage as 'idCOLON000004,same_cov_254_ascending_and_0xa0xa'
            # (also from merge_new_corpus/) neither of those files should be
            # merged into the input corpus.
            '254_ascending_and_0xa0xa',
            'nearly_empty',
            # AFL's dummy file, which was not originally part of the input
            # corpus, but is added by launcher.
            'in1'
        ]),
        sorted(os.listdir(input_corpus)))
    self.assertIn('Merge completed successfully.', self.logged_messages)


class TestLauncher(BaseLauncherTest):
  """AFL launcher tests."""

  def test_abnormal_return_code(self):
    self._test_abnormal_return_code()

  def test_single_testcase_crash(self):
    """Tests launcher with a crashing testcase."""
    testcase_path = setup_testcase_and_corpus('crash', 'empty_corpus')
    output = run_launcher(testcase_path, 'test_fuzzer')
    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address 0x000000000000',
        output)

    # Make sure we didn't fuzz.
    self.assertNotIn('afl-fuzz', output)

  def test_assert(self):
    """Tests launcher with a crashing testcase(assert)."""
    os.environ['ASAN_OPTIONS'] = 'handle_abort=1'
    testcase_path = setup_testcase_and_corpus('crash', 'empty_corpus')
    output = run_launcher(testcase_path, 'assert_fail')
    self.assertIn('Assertion `false\' failed.', output)
    self.assertIn('ERROR: AddressSanitizer: ABRT on unknown address', output)

  @mock.patch('clusterfuzz._internal.bot.fuzzers.afl.launcher.get_fuzz_timeout')
  def test_fuzz_no_crash(self, mock_get_timeout):
    """Tests fuzzing (no crash)."""
    mock_get_timeout.return_value = get_fuzz_timeout(5.0)
    testcase_path = setup_testcase_and_corpus('empty', 'corpus', fuzz=True)
    output = run_launcher(testcase_path, 'test_fuzzer')
    self.assertIn(
        'Command: {0}/afl-fuzz -l2 -pfast -Sdefault -i{1}/corpus '
        '-o{1}/temp-1337/afl_output_dir -mnone '
        '{0}/test_fuzzer 2147483647'.format(DATA_DIRECTORY, TEMP_DIRECTORY),
        output)

    # New items should've been added to the corpus.
    self.assertNotEqual(len(os.listdir(os.environ['FUZZ_CORPUS_DIR'])), 0)

  @unittest.skip('AFL++ does not handle crashes in input corpus properly.')
  @mock.patch('clusterfuzz._internal.bot.fuzzers.afl.launcher.get_fuzz_timeout')
  def test_fuzz_input_crash(self, mock_get_timeout):
    """Tests fuzzing (crash in input)."""
    mock_get_timeout.return_value = get_fuzz_timeout(5.0)
    testcase_path = setup_testcase_and_corpus('empty', 'corpus', fuzz=True)
    output = run_launcher(testcase_path, 'always_crash_fuzzer')
    self.assertIn(
        'Command: {0}/afl-fuzz -l2 -pfast -Sdefault -i{1}/corpus '
        '-o{1}/temp-1337/afl_output_dir -mnone '
        '{0}/always_crash_fuzzer 2147483647'.format(DATA_DIRECTORY,
                                                    TEMP_DIRECTORY), output)
    self.assertIn(
        'ERROR: AddressSanitizer: SEGV on unknown address '
        '0x000000000000', output)

    # No testcase should have been copied back.
    self.assertEqual(os.path.getsize(testcase_path), 0)

  @mock.patch('clusterfuzz._internal.bot.fuzzers.afl.launcher.get_fuzz_timeout')
  def test_fuzz_crash(self, mock_get_timeout):
    """Tests fuzzing (crash)."""
    # *WARNING* Do not lower the fuzz timeout unless you really know what you
    # are doing. Doing so will cause the test to fail on rare ocassion, which
    # will break deploys.
    mock_get_timeout.return_value = get_fuzz_timeout(120.0)
    testcase_path = setup_testcase_and_corpus('empty', 'corpus', fuzz=True)

    output = run_launcher(testcase_path, 'easy_crash_fuzzer')
    self.assertIn(
        'Command: {0}/afl-fuzz -l2 -pfast -Sdefault -i{1}/corpus '
        '-o{1}/temp-1337/afl_output_dir -mnone '
        '{0}/easy_crash_fuzzer 2147483647'.format(DATA_DIRECTORY,
                                                  TEMP_DIRECTORY), output)

    self.assertIn('ERROR: AddressSanitizer: heap-use-after-free on address',
                  output)

    # Testcase should've been copied back.
    self.assertGreaterEqual(os.path.getsize(testcase_path), 3)
    with open(testcase_path, 'rb') as f:
      self.assertEqual(f.read()[:3], b'ABC')

  @no_errors
  @unittest.skip('AFL++ cant consistently find testcases fast enough for test.')
  @mock.patch('clusterfuzz._internal.bot.fuzzers.afl.launcher.get_fuzz_timeout')
  def test_fuzz_merge(self, mock_get_timeout):
    """Tests fuzzing with merge."""
    mock_get_timeout.return_value = get_fuzz_timeout(15.0)
    testcase_path = setup_testcase_and_corpus(
        'empty', 'redundant_corpus', fuzz=True)
    corpus_path = os.environ['FUZZ_CORPUS_DIR']

    for i in range(100):
      with open(os.path.join(corpus_path, '%04d' % i), 'w') as f:
        f.write('A' * 256)
    output = run_launcher(testcase_path, 'test_fuzzer')

    self.assertIn(
        'Command: {0}/afl-fuzz -l2 -pfast -Sdefault -i{1}/redundant_corpus '
        '-o{1}/temp-1337/afl_output_dir -mnone '
        '{0}/test_fuzzer 2147483647'.format(DATA_DIRECTORY, TEMP_DIRECTORY),
        output)

    self.assertIn('Merging corpus.', self.logged_messages)
    self.assertNotIn('Timed out in merge', ' '.join(self.logged_messages))
    self.assertIn('Merge completed successfully.', self.logged_messages)

  @no_errors
  @mock.patch('clusterfuzz._internal.bot.fuzzers.afl.launcher.get_fuzz_timeout')
  def test_libfuzzerize_corpus(self, mock_get_timeout):
    self._test_libfuzzerize_corpus(mock_get_timeout)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""access.py contains static methods around access permissions."""

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_handler
from libs import auth
from libs import helpers
from libs.issue_management import issue_tracker_utils


def _is_privileged_user(email):
  """Check if an email is in the privileged users list."""
  if local_config.AuthConfig().get('all_users_privileged'):
    return True

  privileged_user_emails = (db_config.get_value('privileged_users') or
                            '').splitlines()
  return any(
      utils.emails_equal(email, privileged_user_email)
      for privileged_user_email in privileged_user_emails)


def _is_blacklisted_user(email):
  """Check if an email is in the privileged users list."""
  blacklisted_user_emails = (db_config.get_value('blacklisted_users') or
                             '').splitlines()
  return any(
      utils.emails_equal(email, blacklisted_user_email)
      for blacklisted_user_email in blacklisted_user_emails)


def get_user_job_type():
  """Return the job_type that is assigned to the current user. None means one
    can access any job type. You might want to invoke get_access(..) with
    the job type afterward."""
  email = helpers.get_user_email()
  privileged_user_emails = (db_config.get_value('privileged_users') or
                            '').splitlines()
  for privileged_user_email in privileged_user_emails:
    if ';' in privileged_user_email:
      tokens = privileged_user_email.split(';')
      privileged_user_real_email = tokens[0]
      privileged_user_job_type = tokens[1]
      if utils.emails_equal(email, privileged_user_real_email):
        return privileged_user_job_type
  return None


def _is_domain_allowed(email):
  """Check if the email's domain is allowed."""
  domains = local_config.AuthConfig().get('whitelisted_domains', default=[])
  for domain in domains:
    if utils.normalize_email(email).endswith('@%s' % domain.lower()):
      return True

  return False


class UserAccess(object):
  Allowed, Denied, Redirected = list(range(3))  # pylint: disable=invalid-name


def has_access(need_privileged_access=False, job_type=None, fuzzer_name=None):
  """Check if the user has access."""
  result = get_access(
      need_privileged_access=need_privileged_access,
      job_type=job_type,
      fuzzer_name=fuzzer_name)

  return result == UserAccess.Allowed


def get_access(need_privileged_access=False, job_type=None, fuzzer_name=None):
  """Return 'allowed', 'redirected', or 'failed'."""
  if auth.is_current_user_admin():
    return UserAccess.Allowed

  user = auth.get_current_user()
  if not user:
    return UserAccess.Redirected

  email = user.email
  if _is_blacklisted_user(email):
    return UserAccess.Denied

  if _is_privileged_user(email):
    return UserAccess.Allowed

  if job_type and external_users.is_job_allowed_for_user(email, job_type):
    return UserAccess.Allowed

  if (fuzzer_name and
      external_users.is_fuzzer_allowed_for_user(email, fuzzer_name)):
    return UserAccess.Allowed

  if not need_privileged_access and _is_domain_allowed(email):
    return UserAccess.Allowed

  return UserAccess.Denied


def can_user_access_testcase(testcase):
  """Checks if the current user can access the testcase."""
  config = db_config.get()
  need_privileged_access = (
      testcase.security_flag and not config.relax_security_bug_restrictions)

  if has_access(
      fuzzer_name=testcase.actual_fuzzer_name(),
      job_type=testcase.job_type,
      need_privileged_access=need_privileged_access):
    return True

  user_email = helpers.get_user_email()
  if testcase.uploader_email and testcase.uploader_email == user_email:
    return True

  # Allow owners of bugs to see associated test cases and test case groups.
  issue_id = testcase.bug_information or testcase.group_bug_information
  if not issue_id:
    return False

  issue_tracker = issue_tracker_utils.get_issue_tracker_for_testcase(testcase)
  associated_issue = issue_tracker.get_issue(issue_id)
  if not associated_issue:
    return False

  # Look at both associated issue and original issue (if the associated one
  # is a duplicate of the original issue).
  issues_to_check = [associated_issue]
  if associated_issue.merged_into:
    original_issue = issue_tracker.get_original_issue(issue_id)
    if original_issue:
      issues_to_check.append(original_issue)

  relaxed_restrictions = (
      config.relax_testcase_restrictions or _is_domain_allowed(user_email))
  for issue in issues_to_check:
    if relaxed_restrictions:
      if (any(utils.emails_equal(user_email, cc) for cc in issue.ccs) or
          utils.emails_equal(user_email, issue.assignee) or
          utils.emails_equal(user_email, issue.reporter)):
        return True

    elif utils.emails_equal(user_email, issue.assignee):
      return True

  return False


def check_access_and_get_testcase(testcase_id):
  """Check the failed attempt count and get the testcase."""
  if not helpers.get_user_email():
    raise helpers.UnauthorizedException()

  if not testcase_id:
    raise helpers.EarlyExitException('No test case specified!', 404)

  try:
    testcase = data_handler.get_testcase_by_id(testcase_id)
  except errors.InvalidTestcaseError:
    raise helpers.EarlyExitException('Invalid test case!', 404)

  if not can_user_access_testcase(testcase):
    raise helpers.AccessDeniedException()

  return testcase



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lock functions."""
# NOTE: Deprecated. Avoid using.

import datetime
import random
import time

from google.cloud import ndb
from google.cloud.ndb import exceptions

from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

DEFAULT_MAX_HOLD_SECONDS = 60 * 10

LOCK_CHECK_SLEEP_MULTIPLIER = 10
LOCK_CHECK_TIMEOUT = 60 * 60

MAX_WAIT_EXPONENT = 6
NUM_STATISTICS_SHARDS = 50

TRANSACTION_RETRIES = 0


def _get_current_lock_zone():
  """Get the current zone for locking purposes."""
  if environment.get_value('LOCAL_DEVELOPMENT', False):
    return 'local'

  platform = environment.get_platform_group().lower()
  platform_group_mappings = db_config.get_value('platform_group_mappings')
  for mapping in platform_group_mappings.splitlines():
    if ';' not in mapping:
      continue

    platform_group, zone = mapping.split(';')
    if platform_group.strip() == platform:
      return zone

  # Default to per-platform separation.
  logs.log_warn('Platform group mapping not set in admin configuration, '
                'using default platform - %s.' % platform)
  return platform


def _get_key_name_with_lock_zone(key_name):
  """Return the lock key name with the current lock zone."""
  current_zone = _get_current_lock_zone()
  if not current_zone:
    logs.log_error('Could not find zone.')
    return None

  return current_zone + ';' + key_name


def _try_acquire_lock(key_name, expiration_time, holder):
  """Actual lock acquire that runs in a transaction."""
  lock_entity = ndb.Key(data_types.Lock, key_name).get()

  if lock_entity is None:
    # Lock wasn't held, try to acquire.
    lock_entity = data_types.Lock(
        id=key_name, expiration_time=expiration_time, holder=holder)
    lock_entity.put()
    return lock_entity

  if lock_entity.expiration_time <= datetime.datetime.utcnow():
    # Lock was expired, try to take over the lock.
    lock_entity.expiration_time = expiration_time
    lock_entity.holder = holder
    lock_entity.put()

  return lock_entity


def acquire_lock(key_name,
                 max_hold_seconds=DEFAULT_MAX_HOLD_SECONDS,
                 retries=None,
                 by_zone=True):
  """Acquire a lock for the given key name. Returns the expiration time if
  succeeded, otherwise None. The lock holder is responsible for making sure it
  doesn't assume the lock is still held after the expiration time."""
  logs.log('Acquiring lock for %s.' % key_name)
  failed_acquires = 0
  total_wait = 0
  wait_exponent = 1

  if by_zone:
    key_name_with_zone = _get_key_name_with_lock_zone(key_name)
    if key_name_with_zone is None:
      logs.log_error('Failed to get zone while trying to lock %s.' % key_name)
      return None

    key_name = key_name_with_zone

  bot_name = environment.get_value('BOT_NAME')
  expiration_delta = datetime.timedelta(seconds=max_hold_seconds)
  while total_wait < LOCK_CHECK_TIMEOUT:
    try:
      lock_entity = ndb.transaction(
          lambda: _try_acquire_lock(key_name,
                                    expiration_time=datetime.datetime.utcnow() +
                                    expiration_delta, holder=bot_name),
          retries=TRANSACTION_RETRIES)

      if lock_entity.holder == bot_name:
        logs.log('Got the lock.')
        return lock_entity.expiration_time
    except exceptions.Error:
      pass

    failed_acquires += 1
    if retries and retries >= failed_acquires:
      logs.log('Failed to acquire lock, exceeded max retries.')
      return None

    logs.log('Failed to acquire lock, waiting...')

    # Exponential backoff.
    max_sleep = (1 << wait_exponent) * LOCK_CHECK_SLEEP_MULTIPLIER
    sleep_time = random.uniform(1.0, max_sleep)
    time.sleep(sleep_time)

    total_wait += sleep_time
    wait_exponent = min(wait_exponent + 1, MAX_WAIT_EXPONENT)

  logs.log('Timeout exceeded while trying to acquire lock, bailing.')
  return None


def release_lock(key_name, force_release=False, by_zone=True):
  """Release a lock for the given key name."""
  logs.log('Releasing lock for %s.' % key_name)
  bot_name = environment.get_value('BOT_NAME')

  if by_zone:
    key_name_with_zone = _get_key_name_with_lock_zone(key_name)
    if key_name_with_zone is None:
      logs.log_error('Failed to get zone while releasing %s.' % key_name)
      return

    key_name = key_name_with_zone

  lock_entity = ndb.Key(data_types.Lock, key_name).get()
  if lock_entity and (force_release or lock_entity.holder == bot_name):
    lock_entity.key.delete()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage access to crash. There are 3 union access controls:
  1. Can access everything (maybe without security).
  2. Can access crashes with specific jobs.
  3. Can access crashes with specific fuzzers.

  Crash access control is used by the testcase list page and the crash stats
  page."""

import collections

from clusterfuzz._internal.base import external_users
from clusterfuzz._internal.datastore import data_types
from libs import access
from libs import helpers

Scope = collections.namedtuple('Scope', [
    'everything', 'is_privileged', 'job_types', 'fuzzer_names',
    'allowed_job_type'
])


def get_permission_names(entity_kind):
  """Get scoped fuzzer names."""
  # pylint: disable=protected-access
  permissions = external_users._get_permissions_query_for_user(
      helpers.get_user_email(), entity_kind)

  names = []
  for permission in permissions:
    suffix = '*' if permission.is_prefix else ''
    names.append(permission.entity_name + suffix)
  return names


def get_scope():
  """Get the scope object for the user."""
  user_email = helpers.get_user_email()
  is_privileged = access.has_access(need_privileged_access=True)
  everything = (is_privileged or access.has_access())

  # pylint: disable=protected-access
  job_types = external_users._allowed_entities_for_user(
      user_email, data_types.PermissionEntityKind.JOB)

  allowed_job_type = access.get_user_job_type()
  if allowed_job_type:
    job_types.append(allowed_job_type)

  # pylint: disable=protected-access
  fuzzer_names = external_users._allowed_entities_for_user(
      user_email, data_types.PermissionEntityKind.FUZZER)

  return Scope(everything, is_privileged, job_types, fuzzer_names,
               allowed_job_type)


def add_permissions_to_params(scope, params):
  """Add permissions to params."""
  params['permissions'] = {
      'everything': scope.everything,
      'isPrivileged': scope.is_privileged,
      'jobs': get_permission_names(data_types.PermissionEntityKind.JOB),
      'fuzzers': get_permission_names(data_types.PermissionEntityKind.FUZZER)
  }

  if scope.allowed_job_type:
    params['permissions']['jobs'].append(scope.allowed_job_type)


def add_scope(query, params, security_field, job_type_field, fuzzer_name_field):
  """Add scope to the query according to permissions and modify params."""
  scope = get_scope()
  add_permissions_to_params(scope, params)

  subqueries = []

  if scope.is_privileged:  # The user can access everything.
    return

  if scope.everything:
    everything_query = query.new_subquery()
    everything_query.filter(security_field, False)
    subqueries.append(everything_query)

  if scope.job_types:
    job_query = query.new_subquery()
    job_query.filter_in(job_type_field, scope.job_types)
    subqueries.append(job_query)

  if scope.fuzzer_names:
    fuzzer_query = query.new_subquery()
    fuzzer_query.filter_in(fuzzer_name_field, scope.fuzzer_names)
    subqueries.append(fuzzer_query)

  if not subqueries:  # The user CANNOT access anything.
    raise helpers.AccessDeniedException()

  query.union(*subqueries)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging functions."""

import datetime
import json
import logging
from logging import config
import os
import sys
import time
import traceback

STACKDRIVER_LOG_MESSAGE_LIMIT = 80000  # Allowed log entry size is 100 KB.
LOCAL_LOG_MESSAGE_LIMIT = 100000
LOCAL_LOG_LIMIT = 500000
_logger = None
_is_already_handling_uncaught = False
_default_extras = {}


def _increment_error_count():
  """"Increment the error count metric."""
  if _is_running_on_app_engine():
    task_name = 'appengine'
  else:
    task_name = os.getenv('TASK_NAME', 'unknown')

  from clusterfuzz._internal.metrics import monitoring_metrics
  monitoring_metrics.LOG_ERROR_COUNT.increment({'task_name': task_name})


def _is_local():
  """Return whether or not in a local development environment."""
  return (bool(os.getenv('LOCAL_DEVELOPMENT')) or
          os.getenv('SERVER_SOFTWARE', '').startswith('Development/'))


def _is_running_on_app_engine():
  """Return whether or not we're running on App Engine (production or
  development)."""
  return os.getenv('GAE_ENV') or (
      os.getenv('SERVER_SOFTWARE') and
      (os.getenv('SERVER_SOFTWARE').startswith('Development/') or
       os.getenv('SERVER_SOFTWARE').startswith('Google App Engine/')))


def _console_logging_enabled():
  """Return bool on where console logging is enabled, usually for tests and
  reproduce tool."""
  return bool(os.getenv('LOG_TO_CONSOLE'))


def suppress_unwanted_warnings():
  """Suppress unwanted warnings."""
  # See https://github.com/googleapis/google-api-python-client/issues/299
  logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)


def set_logger(logger):
  """Set the logger."""
  global _logger
  _logger = logger


def get_handler_config(filename, backup_count):
  """Get handler config."""
  root_directory = os.getenv('ROOT_DIR')
  file_path = os.path.join(root_directory, filename)
  max_bytes = 0 if _is_local() else LOCAL_LOG_LIMIT

  return {
      'class': 'logging.handlers.RotatingFileHandler',
      'level': logging.INFO,
      'formatter': 'simple',
      'filename': file_path,
      'maxBytes': max_bytes,
      'backupCount': backup_count,
      'encoding': 'utf8',
  }


def get_logging_config_dict(name):
  """Get config dict for the logger `name`."""
  logging_handler = {
      'run_bot': get_handler_config('bot/logs/bot.log', 3),
      'run': get_handler_config('bot/logs/run.log', 1),
      'run_heartbeat': get_handler_config('bot/logs/run_heartbeat.log', 1),
      'heartbeat': get_handler_config('bot/logs/heartbeat.log', 1),
      'run_fuzzer': get_handler_config('bot/logs/run_fuzzer.log', 1),
      'run_testcase': get_handler_config('bot/logs/run_testcase.log', 1),
  }

  return {
      'version': 1,
      'disable_existing_loggers': False,
      'formatters': {
          'simple': {
              'format': ('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
          }
      },
      'handlers': {
          'handler': logging_handler[name],
          'fluentd': {
              'class': 'clusterfuzz._internal.metrics.logs.JsonSocketHandler',
              'level': logging.INFO,
              'host': '127.0.0.1',
              'port': 5170,
          }
      },
      'loggers': {
          name: {
              'handlers': ['handler']
          }
      },
      'root': {
          'level': logging.INFO,
          'handlers': ['fluentd']
      }
  }


def truncate(msg, limit):
  """We need to truncate the message in the middle if it gets too long."""
  if len(msg) <= limit:
    return msg

  half = limit // 2
  return '\n'.join([
      msg[:half],
      '...%d characters truncated...' % (len(msg) - limit), msg[-half:]
  ])


def format_record(record):
  """Format LogEntry into JSON string."""
  entry = {
      'message':
          truncate(record.getMessage(), STACKDRIVER_LOG_MESSAGE_LIMIT),
      'created': (
          datetime.datetime.utcfromtimestamp(record.created).isoformat() + 'Z'),
      'severity':
          record.levelname,
      'bot_name':
          os.getenv('BOT_NAME'),
      'task_payload':
          os.getenv('TASK_PAYLOAD'),
      'name':
          record.name,
  }

  entry['location'] = getattr(record, 'location', {'error': True})
  entry['extras'] = getattr(record, 'extras', {})
  update_entry_with_exc(entry, record.exc_info)

  if not entry['extras']:
    del entry['extras']

  worker_bot_name = os.environ.get('WORKER_BOT_NAME')
  if worker_bot_name:
    entry['worker_bot_name'] = worker_bot_name

  fuzz_target = os.getenv('FUZZ_TARGET')
  if fuzz_target:
    entry['fuzz_target'] = fuzz_target

  # Log bot shutdown cases as WARNINGs since this is expected for preemptibles.
  if (entry['severity'] in ['ERROR', 'CRITICAL'] and
      'IOError: [Errno 4] Interrupted function call' in entry['message']):
    entry['severity'] = 'WARNING'

  return json.dumps(entry)


def update_entry_with_exc(entry, exc_info):
  """Update the dict `entry` with exc_info."""
  if not exc_info:
    return

  error = exc_info[1]
  error_extras = getattr(error, 'extras', {})
  entry['task_payload'] = (
      entry.get('task_payload') or error_extras.pop('task_payload', None))
  entry['extras'].update(error_extras)
  entry['serviceContext'] = {'service': 'bots'}

  # Reference:
  # https://cloud.google.com/error-reporting/docs/formatting-error-messages,
  if exc_info[0]:
    # we need to set the result of traceback.format_exception to the field
    # `message`. And we move our
    entry['message'] += '\n' + ''.join(
        traceback.format_exception(exc_info[0], exc_info[1], exc_info[2]))
  else:
    # If we log error without exception, we need to set
    # `context.reportLocation`.
    location = entry.get('location', {})
    entry['context'] = {
        'reportLocation': {
            'filePath': location.get('path', ''),
            'lineNumber': location.get('line', 0),
            'functionName': location.get('method', '')
        }
    }


class JsonSocketHandler(logging.handlers.SocketHandler):
  """Format log into JSON string before sending it to fluentd. We need this
    because SocketHandler doesn't respect the formatter attribute."""

  def makePickle(self, record):
    """Format LogEntry into JSON string."""
    # \n is the recognized delimiter by fluentd's in_tcp. Don't remove.
    return (format_record(record) + '\n').encode('utf-8')


def uncaught_exception_handler(exception_type, exception_value,
                               exception_traceback):
  """Handles any exception that are uncaught by logging an error and calling
  the sys.__excepthook__."""
  # Ensure that we are not calling ourself. This shouldn't be needed since we
  # are using sys.__excepthook__. Do this check anyway since if we are somehow
  # calling ourself we might infinitely send errors to the logs, which would be
  # quite bad.
  global _is_already_handling_uncaught
  if _is_already_handling_uncaught:
    raise Exception('Loop in uncaught_exception_handler')
  _is_already_handling_uncaught = True

  # Use emit since log_error needs sys.exc_info() to return this function's
  # arguments to call init properly.
  # Don't worry about emit() throwing an Exception, python will let us know
  # about that exception as well as the original one.
  emit(
      logging.ERROR,
      'Uncaught exception',
      exc_info=(exception_type, exception_value, exception_traceback))

  sys.__excepthook__(exception_type, exception_value, exception_traceback)


def configure_appengine():
  """Configure logging for App Engine."""
  logging.getLogger().setLevel(logging.INFO)

  if os.getenv('LOCAL_DEVELOPMENT') or os.getenv('PY_UNITTESTS'):
    return

  import google.cloud.logging
  client = google.cloud.logging.Client()
  handler = client.get_default_handler()
  logging.getLogger().addHandler(handler)


def configure(name, extras=None):
  """Set logger. See the list of loggers in bot/config/logging.yaml.
  Also configures the process to log any uncaught exceptions as an error.
  |extras| will be included by emit() in log messages."""
  suppress_unwanted_warnings()

  if _is_running_on_app_engine():
    configure_appengine()
    return

  if _console_logging_enabled():
    logging.basicConfig()
  else:
    config.dictConfig(get_logging_config_dict(name))

  logger = logging.getLogger(name)
  logger.setLevel(logging.INFO)
  set_logger(logger)

  # Set _default_extras so they can be used later.
  if extras is None:
    extras = {}
  global _default_extras
  _default_extras = extras

  # Install an exception handler that will log an error when there is an
  # uncaught exception.
  sys.excepthook = uncaught_exception_handler


def get_logger():
  """Return logger. We need this method because we need to mock logger."""
  if _logger:
    return _logger

  if _is_running_on_app_engine():
    # Running on App Engine.
    set_logger(logging.getLogger())

  elif _console_logging_enabled():
    # Force a logger when console logging is enabled.
    configure('root')

  return _logger


def get_source_location():
  """Return the caller file, lineno, and funcName."""
  try:
    raise Exception()
  except:
    # f_back is called twice. Once to leave get_source_location(..) and another
    # to leave emit(..).
    # The code is adapted from:
    # https://github.com/python/cpython/blob/2.7/Lib/logging/__init__.py#L1244
    frame = sys.exc_info()[2].tb_frame.f_back

    while frame and hasattr(frame, 'f_code'):
      if not frame.f_code.co_filename.endswith('logs.py'):
        return frame.f_code.co_filename, frame.f_lineno, frame.f_code.co_name
      frame = frame.f_back

  return 'Unknown', '-1', 'Unknown'


def _add_appengine_trace(extras):
  """Add App Engine tracing information."""
  if not _is_running_on_app_engine():
    return

  from libs import auth

  try:
    request = auth.get_current_request()
    if not request:
      return
  except Exception:
    # FIXME: Find a way to add traces in threads. Skip adding for now, as
    # otherwise, we hit an exception "Request global variable is not set".
    return

  trace_header = request.headers.get('X-Cloud-Trace-Context')
  if not trace_header:
    return

  project_id = os.getenv('APPLICATION_ID')
  trace_id = trace_header.split('/')[0]
  extras['logging.googleapis.com/trace'] = (
      'projects/{project_id}/traces/{trace_id}').format(
          project_id=project_id, trace_id=trace_id)


def emit(level, message, exc_info=None, **extras):
  """Log in JSON."""
  logger = get_logger()
  if not logger:
    return

  # Include extras passed as an argument and default extras.
  all_extras = _default_extras.copy()
  all_extras.update(extras)

  path_name, line_number, method_name = get_source_location()

  if _is_running_on_app_engine():
    if exc_info == (None, None, None):
      # Don't pass exc_info at all, as otherwise cloud logging will append
      # "NoneType: None" to the message.
      exc_info = None

    if level >= logging.ERROR:
      # App Engine only reports errors if there is an exception stacktrace, so
      # we generate one. We don't create an exception here and then format it,
      # as that will not include frames below this emit() call. We do [:-2] on
      # the stacktrace to exclude emit() and the logging function below it (e.g.
      # log_error).
      message = (
          message + '\n' + 'Traceback (most recent call last):\n' + ''.join(
              traceback.format_stack()[:-2]) + 'LogError: ' + message)

    _add_appengine_trace(all_extras)

  # We need to make a dict out of it because member of the dict becomes the
  # first class attributes of LogEntry. It is very tricky to identify the extra
  # attributes. Therefore, we wrap extra fields under the attribute 'extras'.
  logger.log(
      level,
      truncate(message, LOCAL_LOG_MESSAGE_LIMIT),
      exc_info=exc_info,
      extra={
          'extras': all_extras,
          'location': {
              'path': path_name,
              'line': line_number,
              'method': method_name
          }
      })


def log(message, level=logging.INFO, **extras):
  """Logs the message to a given log file."""
  emit(level, message, **extras)


def log_warn(message, **extras):
  """Logs the warning message."""
  emit(logging.WARN, message, exc_info=sys.exc_info(), **extras)


def log_error(message, **extras):
  """Logs the error in the error log file."""
  exception = extras.pop('exception', None)
  if exception:
    try:
      raise exception
    except:
      emit(logging.ERROR, message, exc_info=sys.exc_info(), **extras)
  else:
    emit(logging.ERROR, message, exc_info=sys.exc_info(), **extras)
  _increment_error_count()


def log_fatal_and_exit(message, **extras):
  """Logs a fatal error and exits."""
  wait_before_exit = extras.pop('wait_before_exit', None)
  emit(logging.CRITICAL, message, exc_info=sys.exc_info(), **extras)
  _increment_error_count()
  if wait_before_exit:
    log('Waiting for %d seconds before exit.' % wait_before_exit)
    time.sleep(wait_before_exit)
  sys.exit(-1)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for CSP."""

import unittest

from libs import csp


class CSPBuilderTest(unittest.TestCase):
  """Tests for CSPBuilder."""

  def test_simple_policy(self):
    """Ensure that generating a simple policy works as expected."""
    builder = csp.CSPBuilder()

    builder.add('default-src', 'none', quote=True)
    builder.add('connect-src', 'self', quote=True)
    builder.add('script-src', 'self', quote=True)
    builder.add('script-src', 'scripts.test.tld')
    builder.add_sourceless('upgrade-insecure-requests')

    self.assertEqual(
        str(builder), "connect-src 'self'; default-src 'none'; "
        "script-src 'self' scripts.test.tld; upgrade-insecure-requests;")

  def test_policy_modification(self):
    """Ensure that policies can be modified."""
    builder = csp.CSPBuilder()

    builder.add('default-src', 'none', quote=True)
    builder.add('script-src', 'self', quote=True)
    builder.add('script-src', 'unsafe-inline', quote=True)
    builder.add('script-src', 'external.site')
    self.assertEqual(
        str(builder),
        "default-src 'none'; script-src 'self' 'unsafe-inline' external.site;")

    builder.remove('script-src', 'unsafe-inline', quote=True)
    builder.remove('script-src', 'external.site')
    self.assertEqual(str(builder), "default-src 'none'; script-src 'self';")

  def test_exception_on_duplicate_add(self):
    """Ensure that an exception is thrown if we add a duplicate item."""
    builder = csp.CSPBuilder()
    builder.add('default-src', 'none', quote=True)

    with self.assertRaises(AssertionError):
      builder.add('default-src', 'none', quote=True)

    with self.assertRaises(AssertionError):
      builder.add_sourceless('default-src')

    builder.add_sourceless('block-all-mixed-content')
    with self.assertRaises(AssertionError):
      builder.add_sourceless('block-all-mixed-content')

  def test_exception_on_bad_removal(self):
    """Ensure that an exception is thrown if we remove a nonexistent item."""
    builder = csp.CSPBuilder()

    builder.add('default-src', 'none', quote=True)
    builder.remove('default-src', 'none', quote=True)

    with self.assertRaises(AssertionError):
      builder.remove('default-src', 'none', quote=True)

    with self.assertRaises(AssertionError):
      builder.remove('script-src', 'unadded.domain')

  def test_no_exception_from_default_policy(self):
    """Ensure that no exceptions are raised building the default policy."""
    csp.get_default()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for creating issue."""

from flask import request

from handlers import base_handler
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Handler that creates an issue."""

  @staticmethod
  def delete_testcase(testcase_id):
    """Delete a testcase."""
    testcase = helpers.get_testcase(testcase_id)

    # Don't delete testcases that have an associated issue.
    if testcase.bug_information:
      raise helpers.EarlyExitException(
          'The testcase (id=%d) with an assigned issue cannot be deleted.' %
          testcase_id, 400)

    testcase.key.delete()
    helpers.log('Deleted testcase %s' % testcase_id, helpers.MODIFY_OPERATION)

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_admin_access
  def post(self):
    """Delete a testcase."""
    testcase_id = request.get('testcaseId')
    return self.render_json({'testcaseId': self.delete_testcase(testcase_id)})



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""datastore_query tests."""
# pylint: disable=protected-access
import datetime
import unittest

from google.cloud import ndb
import mock
import six

from clusterfuzz._internal.tests.test_libs import test_utils
from libs.query import datastore_query


class TestDatastoreModel(ndb.Model):
  tokens = ndb.StringProperty(repeated=True)
  boolean_value = ndb.BooleanProperty()
  datetime_value = ndb.DateTimeProperty()


def _create_data(self):
  """Create test data."""
  self.mocks = []
  for i in range(30):
    m = TestDatastoreModel()
    if (i % 2) == 0:
      m.tokens = ['a']
      m.boolean_value = True
    else:
      m.tokens = ['b']
      m.boolean_value = False
    m.datetime_value = datetime.datetime.fromtimestamp(100 - i)
    m.put()
    self.mocks.append(m)


@test_utils.with_cloud_emulators('datastore')
class KeyQueryTest(unittest.TestCase):
  """Test KeyQuery."""

  def setUp(self):
    _create_data(self)

  def test_empty(self):
    """Test when there's no record."""
    query = datastore_query._KeyQuery(TestDatastoreModel)
    query.filter('=', 'tokens', 'c')
    query.order('datetime_value', True)

    items, count, has_more = query.fetch(offset=0, limit=10, more_limit=20)
    self.assertListEqual([], items)
    self.assertEqual(0, count)
    self.assertFalse(has_more)

  def test_no_or_condition(self):
    """Test no OR condition."""
    query = datastore_query._KeyQuery(TestDatastoreModel)
    query.filter('=', 'tokens', 'a')
    query.order('datetime_value', True)

    items, count, has_more = query.fetch(offset=8, limit=2, more_limit=10)
    self.assertListEqual([self.mocks[16].key.id(), self.mocks[18].key.id()],
                         [item.key.id() for item in items])
    self.assertEqual(15, count)
    self.assertFalse(has_more)

  def test_or_condition(self):
    """Test OR conditions."""
    query = datastore_query._KeyQuery(TestDatastoreModel)
    query.filter('IN', 'tokens', ['a', 'b', 'c'])
    query.filter('IN', 'boolean_value', [True, False])
    query.order('datetime_value', True)

    queries = query.flatten()
    for q in queries:
      self.assertListEqual([], q.or_filters)
      self.assertEqual('datetime_value', q.order_property)
      self.assertTrue(q.order_desc)

    def _make(token, boolean):
      return [('=', 'tokens', token), ('=', 'boolean_value', boolean)]

    self.assertEqual(6, len(queries))
    self.assertListEqual(_make('a', True), queries[0].filters)
    self.assertListEqual(_make('b', True), queries[1].filters)
    self.assertListEqual(_make('c', True), queries[2].filters)
    self.assertListEqual(_make('a', False), queries[3].filters)
    self.assertListEqual(_make('b', False), queries[4].filters)
    self.assertListEqual(_make('c', False), queries[5].filters)

    items, count, has_more = query.fetch(offset=8, limit=2, more_limit=40)
    self.assertListEqual([self.mocks[8].key.id(), self.mocks[9].key.id()],
                         [item.key.id() for item in items])
    self.assertEqual(30, count)
    self.assertFalse(has_more)

  def test_get_more(self):
    """Test multiple OR conditions and get more items for total count."""
    query = datastore_query._KeyQuery(TestDatastoreModel)
    query.filter('IN', 'tokens', ['a', 'b'])
    query.order('datetime_value', True)

    items, count, has_more = query.fetch(offset=8, limit=2, more_limit=15)
    self.assertListEqual([self.mocks[8].key.id(), self.mocks[9].key.id()],
                         [item.key.id() for item in items])
    self.assertEqual(25, count)
    self.assertTrue(has_more)

  def test_has_more_but_not_get_more(self):
    """Test multiple OR conditions and compute total count from the current
      result."""
    query = datastore_query._KeyQuery(TestDatastoreModel)
    query.filter('IN', 'tokens', ['a', 'b'])
    query.order('datetime_value', True)

    items, count, has_more = query.fetch(offset=8, limit=2, more_limit=5)
    self.assertListEqual([self.mocks[8].key.id(), self.mocks[9].key.id()],
                         [item.key.id() for item in items])
    self.assertEqual(15, count)
    self.assertTrue(has_more)


@test_utils.with_cloud_emulators('datastore')
class QueryTest(unittest.TestCase):
  """Test Query."""

  def setUp(self):
    _create_data(self)

  def test_third_page(self):
    """Test getting the third page with more total count."""
    query = datastore_query.Query(TestDatastoreModel)
    query.filter_in('tokens', ['a', 'b'])
    query.filter('boolean_value', True)
    query.order('datetime_value', is_desc=True)

    items, total_pages, total_items, has_more = query.fetch_page(
        page=3, page_size=2, projection=['tokens'], more_limit=4)
    self.assertListEqual([self.mocks[8].key.id(), self.mocks[10].key.id()],
                         [item.key.id() for item in items])
    self.assertListEqual([['a'], ['a']], [item.tokens for item in items])
    with self.assertRaises(ndb.UnprojectedPropertyError):
      _ = [item.boolean_value for item in items]
    self.assertEqual(10, total_items)
    self.assertEqual(5, total_pages)
    self.assertTrue(has_more)

  def test_greater_or_equal(self):
    """Test that a query using an operator other than "=" works."""
    query = datastore_query.Query(TestDatastoreModel)
    query.filter(
        'datetime_value', datetime.datetime.fromtimestamp(96), operator='>=')
    query.order('datetime_value', is_desc=True)
    _, total_count, has_more = query.fetch(
        offset=0, limit=100, projection=['tokens'], more_limit=100)

    # We expect the above query to return values from 96-100.
    self.assertEqual(total_count, 5)
    self.assertFalse(has_more)

  def test_negative_page(self):
    """Test getting a negative page."""
    query = datastore_query.Query(TestDatastoreModel)
    query.filter_in('tokens', ['a', 'b'])
    query.filter('boolean_value', True)
    query.order('datetime_value', is_desc=True)

    items, total_pages, total_items, has_more = query.fetch_page(
        page=-5, page_size=2, projection=['tokens'], more_limit=4)
    self.assertListEqual([self.mocks[0].key.id(), self.mocks[2].key.id()],
                         [item.key.id() for item in items])
    self.assertListEqual([['a'], ['a']], [item.tokens for item in items])
    with self.assertRaises(ndb.UnprojectedPropertyError):
      _ = [item.boolean_value for item in items]
    self.assertEqual(6, total_items)
    self.assertEqual(3, total_pages)
    self.assertTrue(has_more)


class QueryWrapper(ndb.Query):
  """Query wrapper for easy mocking."""

  def __init__(self, wrapped, results, subqueries):  # pylint: disable=super-init-not-called
    self.wrapped = wrapped
    self.results = results
    self.subqueries = subqueries

  def filter(self, *args):
    """Wraps the result from filter()."""
    query = QueryWrapper(
        ndb.Query.filter(self.wrapped, *args), self.results, self.subqueries)
    self.subqueries.append(query)
    return query

  def order(self, *args):
    """Wraps the result from order()."""
    query = QueryWrapper(
        ndb.Query.order(self.wrapped, *args), self.results, self.subqueries)
    self.subqueries.append(query)
    return query

  def fetch(self, limit=None, **kwargs):  # pylint: disable=unused-argument
    """Wraps fetch()."""
    return self.results

  def iter(self, **kwargs):  # pylint: disable=unused-argument
    m = mock.MagicMock()
    m.__iter__.return_value = self.results
    m.cursor_after.return_value = None
    return m

  def __getattr__(self, attr):
    """Forward all other getters to the actual ndb.Query."""
    return getattr(self.wrapped, attr)


@test_utils.with_cloud_emulators('datastore')
class QueryMockTest(unittest.TestCase):
  """Test Query with mocks. This test is important because we want to make sure
    we call the underlying query correctly."""

  def setUp(self):
    original_query = TestDatastoreModel.query

    def get_query(*args, **kwargs):
      """Mock query."""
      query = original_query(*args, **kwargs)
      item = mock.MagicMock()
      item.key = ndb.Key(TestDatastoreModel, len(self.queries))
      item.datetime_value = datetime.datetime.utcnow()

      self.queries.append([query])
      return QueryWrapper(query, [item], self.queries[-1])

    self.queries = []
    patcher = mock.patch.object(TestDatastoreModel, 'query')
    mock_query = patcher.start()
    mock_query.side_effect = get_query
    self.addCleanup(patcher.stop)

  def test_third_page(self):
    """Test getting the third page with more total count."""
    query = datastore_query.Query(TestDatastoreModel)
    query.filter_in('tokens', ['a', 'b'])
    query.filter('boolean_value', True)
    query.order('datetime_value', is_desc=True)

    query.fetch_page(page=1, page_size=2, projection=['tokens'], more_limit=4)

    self.assertIsInstance(self.queries[0][-1].filters, ndb.AND)
    six.assertCountEqual(self, [
        ('tokens', '=', 'a'),
        ('boolean_value', '=', True),
    ], [f.__getnewargs__() for f in self.queries[0][-1].filters])

    self.assertIsInstance(self.queries[1][-1].filters, ndb.AND)
    six.assertCountEqual(self, [
        ('tokens', '=', 'b'),
        ('boolean_value', '=', True),
    ], [f.__getnewargs__() for f in self.queries[1][-1].filters])

    self.assertIsInstance(self.queries[2][-1].filters, ndb.OR)

    expected = []
    for item in [f.__getnewargs__() for f in self.queries[2][-1].filters]:
      expected.append((item[0], item[1], repr(item[2])))

    six.assertCountEqual(self, [
        ('__key__', '=',
         '<Key(\'TestDatastoreModel\', 0), project=test-clusterfuzz>'),
        ('__key__', '=',
         '<Key(\'TestDatastoreModel\', 1), project=test-clusterfuzz>'),
    ], expected)


class ComputeProjectionTest(unittest.TestCase):
  """Test compute_projection."""

  def test_none(self):
    """Test when projection is None."""
    self.assertIsNone(datastore_query.compute_projection(None, 'order_field'))

  def test_combine(self):
    """Test combine."""
    self.assertSetEqual(
        set(['a', 'b', 'c']),
        set(datastore_query.compute_projection(['a', 'c'], 'b')))

  def test_dedup(self):
    """Test dedup."""
    self.assertSetEqual(
        set(['a', 'b', 'c']),
        set(datastore_query.compute_projection(['a', 'b', 'c'], 'b')))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Allow users to configure the reproduce tool to point to this site."""

import urllib.parse

from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.datastore import data_handler
from handlers import base_handler
from libs import handler


class Handler(base_handler.Handler):
  """Handler to configure the reproduce tool."""

  # Note: This handler is intentionally unauthenticated.
  @handler.post(handler.JSON, handler.JSON)
  def post(self):
    """Download the reproduce tool configuration json."""
    client_id = db_config.get_value('reproduce_tool_client_id')
    if not client_id:
      return self.render_json({
          'error': 'Reproduce tool is not configured.'
      }, 500)

    domain = data_handler.get_domain()
    link_format = 'https://{domain}/{handler}'
    configuration = {
        'testcase_info_url':
            link_format.format(
                domain=domain, handler='reproduce-tool/testcase-info'),
        'testcase_download_url':
            link_format.format(
                domain=domain, handler='testcase-detail/download-testcase'),
        'oauth_url':
            'https://accounts.google.com/o/oauth2/v2/auth?{}'.format(
                urllib.parse.urlencode({
                    'client_id': client_id,
                    'scope': 'email profile',
                    'response_type': 'code',
                    'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob'
                })),
    }

    return self.render_json(configuration)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Statistics for the afl launcher script."""

import os
import re

import six

from clusterfuzz._internal.bot.fuzzers import dictionary_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

SANITIZER_START_REGEX = re.compile(r'.*ERROR: [A-z]+Sanitizer:.*')
SANITIZER_SEPERATOR_REGEX = re.compile(r'^=+$')


class StatsGetter(object):
  """Calculates stats and retreives stats, compiling them into dictionary that
  can be uploaded to ClusterFuzz.
  """

  # Regex for finding stat names and values in fuzzer_stats and
  # extra_fuzzer_stats.
  STATS_REGEX = r'(?P<stat_key>\S+)\s+: (?P<stat_val>\d+[.\d]*).*\n'

  # Mapping from names in AFL's stats file to those of Clusterfuzz's
  AFL_STATS_MAPPING = {
      'exec_timeout': 'timeout_limit',
      'stability': 'stability',
      'unique_crashes': 'crash_count',
      'unique_hangs': 'timeout_count',
  }

  NO_INSTRUMENTATION_REGEX = re.compile(
      r'PROGRAM ABORT :.*No instrumentation detected')
  STARTUP_CRASH_REGEX = re.compile(r'target binary (crashed|terminated)')
  CORPUS_CRASH_REGEX = re.compile(
      r'program crashed with one of the test cases provided')

  def __init__(self, afl_stats_path, dict_path):
    """Set important attributes and initialize all stats to 0.

    Args:
      afl_stats_path: The path to the fuzzer_stats file afl-fuzz outputs.
      Note that we do not use the term "fuzzer_stats" in this parameter and
      associated vairables so that they won't be confused with the module
      "fuzzer_stats".
      dict_path: The path to the dictionary file given to afl-fuzz as input.
    """
    self.afl_stats_path = afl_stats_path

    # The parsed stats in afl_stats_path.
    self.afl_stats = {}
    self.dict_path = dict_path

    # Default values.
    self.stats = {
        'actual_duration': 0,
        'average_exec_per_sec': 0,
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'crash_count': 0,
        'dict_used': 0,
        'log_lines_unwanted': 0,
        'manual_dict_size': 0,
        'new_units_added': 0,
        'new_units_generated': 0,
        'stability': 0.0,
        'startup_crash_count': 0,
        'timeout_count': 0,
        'timeout_limit': 0,
    }

  def set_afl_stats(self):
    """Read statistics from afl-fuzz's "fuzzer_stats" file and save them as
    self.afl_stats.
    """
    # fuzzer_stats will not exist if there was a crashing input in the corpus or
    # if afl was unable to fuzz for some other reason.
    if not os.path.exists(self.afl_stats_path):
      return

    afl_stats_string = engine_common.read_data_from_file(
        self.afl_stats_path).decode('utf-8')
    matches_iterator = re.finditer(self.STATS_REGEX, afl_stats_string)
    self.afl_stats = dict(match.groups() for match in matches_iterator)

  def get_afl_stat(self, afl_stat_name):
    """Try to get |afl_stat_name| from self.afl_stats, otherwise return a
    default value instead. Note that is imporant that the types are correct
    since a type change will prevent this data from being loaded into
    BigQuery.

      Args:
        afl_stat_name: The name of the stat from the fuzzer_stats file.

      Returns:
        The int value of |afl_stat| in fuzzer_stats if it is an int, or the
        float value if it is a float. If the value is not in fuzzer_stats, then
        return its default value if one was specified, or zero if none was
        specified.
    """
    # If the stat isn't in afl's stats file, then use the default value
    # specified.
    try:
      afl_stat_value = self.afl_stats[afl_stat_name]
    except KeyError:
      print("{0} not in AFL's stats file.".format(afl_stat_name))
      # If afl_stat_value is in AFL_STATS_MAPPING, then get the clusterfuzz name
      # of the stat to lookup the stat's default value.
      clusterfuzz_stat_name = self.AFL_STATS_MAPPING.get(
          afl_stat_name, afl_stat_name)

      return self.stats.get(clusterfuzz_stat_name, 0)

    # If we find the stat in the file, try returning it as an int. If it is a
    # float, that will fail, so return it as a float.
    try:
      return int(afl_stat_value)
    except ValueError:
      return float(afl_stat_value)

  def _get_unwanted_log_line_count(self, log_output):
    """Return number of unwanted lines in log output."""
    count = 0
    for line in log_output.splitlines():
      if SANITIZER_SEPERATOR_REGEX.match(line):
        continue
      if SANITIZER_START_REGEX.match(line):
        break
      count += 1

    return count

  def set_stats(self,
                actual_duration,
                new_units_generated,
                new_units_added,
                corpus_size,
                fuzzing_strategies,
                fuzzer_stderr,
                afl_fuzz_output=''):
    """Create a dict of statistics that can be uploaded to ClusterFuzz and save
    it in self.stats.

    Args:
      actual_duration: The length of time afl-fuzz was run for, in seconds.
      new_units_generated: The number of new corpus files generated by afl-fuzz.
      new_units_added: The number of new corpus files left after minimizing
      those generated by afl-fuzz.
      fuzzing_strategies: Fuzzing strategies used by AFL.
      afl_fuzz_output: Output from afl-fuzz.

    Returns:
      The stats dictionary. Any values that could not be found default to 0.
    """
    # TODO(metzman): Add expected_duration to stats.
    # TODO(metzman): Add the other stats that are less clear how to add.

    assert actual_duration >= 0
    assert new_units_generated >= 0
    assert new_units_added >= 0
    assert new_units_added <= new_units_generated

    # Set stats passed to this function as arguments.
    self.stats['actual_duration'] = int(actual_duration)
    self.stats['new_units_generated'] = new_units_generated
    self.stats['new_units_added'] = new_units_added
    self.stats['corpus_size'] = corpus_size

    # Set log_lines_unwanted stat from parsing fuzzer stderr.
    self.stats['log_lines_unwanted'] = self._get_unwanted_log_line_count(
        fuzzer_stderr)

    # Set dictionary stats if self.dict_path is set.
    if self.dict_path is not None:
      self.stats['dict_used'] = 1
      self.stats['manual_dict_size'], _ = (
          dictionary_manager.get_stats_for_dictionary_file(self.dict_path))

    # Read and parse stats from AFL's afl_stats. Then use them to set and
    # calculate our own stats.
    self.set_afl_stats()
    for afl_stat, clusterfuzz_stat in six.iteritems(self.AFL_STATS_MAPPING):
      self.stats[clusterfuzz_stat] = self.get_afl_stat(afl_stat)

    try:
      self.stats['average_exec_per_sec'] = int(
          self.get_afl_stat('execs_done') // actual_duration)

    except ZeroDivisionError:  # Fail gracefully if actual_duration is 0.
      self.stats['average_exec_per_sec'] = 0
      logs.log_error('actual_duration is 0 in fuzzer_stats. '
                     'average_exec_per_sec defaulting to 0.')

    # Normalize |timeout_count| and |crash_count| to be either 0 or 1.
    for stat_variable in ['crash_count', 'timeout_count']:
      self.stats[stat_variable] = int(bool(self.stats[stat_variable]))

    self.set_strategy_stats(fuzzing_strategies)
    self.set_output_stats(afl_fuzz_output)
    return self.stats

  def set_output_stats(self, afl_fuzz_output):
    """Set stats gotten from the output of afl-fuzz, |afl_fuzz_output|."""
    # If there is no instrumentation, note it.
    if re.search(self.NO_INSTRUMENTATION_REGEX, afl_fuzz_output):
      self.stats['bad_instrumentation'] = 1

    # If there is a startup crash, note it.
    if re.search(self.STARTUP_CRASH_REGEX, afl_fuzz_output):
      self.stats['startup_crash_count'] = 1

    # If there is a crashing input in corpus, note it.
    if re.search(self.CORPUS_CRASH_REGEX, afl_fuzz_output):
      self.stats['corpus_crash_count'] = 1

  def set_strategy_stats(self, fuzzing_strategies):
    """Sets strategy related stats for afl-fuzz to correct values based on
    |strategies|."""

    self.stats['strategy_selection_method'] = environment.get_value(
        'STRATEGY_SELECTION_METHOD', default_value='default')

    if fuzzing_strategies.use_corpus_subset:
      self.stats['strategy_' + strategy.CORPUS_SUBSET_STRATEGY.name] = (
          fuzzing_strategies.corpus_subset_size)

    if (fuzzing_strategies.generator_strategy == engine_common.Generator.RADAMSA
       ):
      self.stats['strategy_' +
                 strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.name] = 1
    elif (fuzzing_strategies.generator_strategy ==
          engine_common.Generator.ML_RNN):
      self.stats['strategy_' +
                 strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.name] = 1



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get values / settings from local configuration."""

import os

import six
import yaml

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.system import environment

CACHE_SIZE = 1024
YAML_FILE_EXTENSION = '.yaml'

SEPARATOR = '.'

GAE_AUTH_PATH = 'gae.auth'
GAE_CONFIG_PATH = 'gae.config'
GCE_CLUSTERS_PATH = 'gce.clusters'
ISSUE_TRACKERS_PATH = 'issue_trackers.config'
MONITORING_REGIONS_PATH = 'monitoring.regions'
PROJECT_PATH = 'project'


def _load_yaml_file(yaml_file_path):
  """Load yaml file and return parsed contents."""
  with open(yaml_file_path) as f:
    try:
      return yaml.safe_load(f.read())
    except Exception:
      raise errors.ConfigParseError(yaml_file_path)


def _find_key_in_yaml_file(yaml_file_path, search_keys, full_key_name,
                           value_is_relative_path):
  """Find a key in a yaml file."""
  if not os.path.isfile(yaml_file_path):
    return None

  result = _load_yaml_file(yaml_file_path)

  if not search_keys:
    # Give the entire yaml file contents.
    # |value_is_relative_path| is not applicable here.
    return result

  for search_key in search_keys:
    if not isinstance(result, dict):
      raise errors.InvalidConfigKey(full_key_name)

    if search_key not in result:
      return None

    result = result[search_key]

  if value_is_relative_path:
    yaml_directory = os.path.dirname(yaml_file_path)
    if isinstance(result, list):
      result = [os.path.join(yaml_directory, str(i)) for i in result]
    else:
      result = os.path.join(yaml_directory, str(result))

  return result


def _get_key_location(search_path, full_key_name):
  """Get the path of the the yaml file and the key components given a full key
  name."""
  key_parts = full_key_name.split(SEPARATOR)
  dir_path = search_path

  # Find the directory components of the key path
  for i, search_key in enumerate(key_parts):
    search_path = os.path.join(dir_path, search_key)
    if os.path.isdir(search_path):
      # Don't allow both a/b/... and a/b.yaml
      if os.path.isfile(search_path + YAML_FILE_EXTENSION):
        raise errors.InvalidConfigKey(full_key_name)

      dir_path = search_path
    else:
      # The remainder of the key path is a yaml_filename.key1.key2...
      key_parts = key_parts[i:]
      break
  else:
    # The entirety of the key reference a directory.
    key_parts = []

  if key_parts:
    return dir_path, key_parts[0] + YAML_FILE_EXTENSION, key_parts[1:]

  return dir_path, '', []


def _validate_root(search_path, root):
  """Validate that a root is valid."""
  if root is None:
    return True

  directory, filename, search_keys = _get_key_location(search_path, root)

  if not filename:
    # _get_key_location already validated that the directory exists, so the root
    # is valid.
    return True

  # Check that the yaml file and keys exist.
  yaml_path = os.path.join(directory, filename)
  return (_find_key_in_yaml_file(
      yaml_path, search_keys, root, value_is_relative_path=False) is not None)


def _search_key(search_path, full_key_name, value_is_relative_path):
  """Search the key in a search path."""
  directory, filename, search_keys = _get_key_location(search_path,
                                                       full_key_name)

  # Search in the yaml file.
  yaml_path = os.path.join(directory, filename)
  return _find_key_in_yaml_file(yaml_path, search_keys, full_key_name,
                                value_is_relative_path)


class Config(object):
  """Config class helper."""

  def __init__(self, root=None, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
    self._root = root.format(*args, **kwargs) if root is not None else None
    self._config_dir = environment.get_config_directory()
    self._cache = memoize.FifoInMemory(CACHE_SIZE)

    # Check that config directory is valid.
    if not self._config_dir or not os.path.exists(self._config_dir):
      raise errors.BadConfigError(self._config_dir)

    # Config roots should exist.
    if not _validate_root(self._config_dir, self._root):
      raise errors.BadConfigError(self._config_dir)

  def sub_config(self, path):
    """Return a new config with a new sub-root."""
    if self._root:
      new_root = self._root + SEPARATOR + path
    else:
      new_root = path

    return Config(root=new_root)

  def _get_helper(self, key_name='', default=None,
                  value_is_relative_path=False):
    """Helper for get and get_absolute_functions."""
    if self._root:
      key_name = self._root + SEPARATOR + key_name if key_name else self._root

    if not key_name:
      raise errors.InvalidConfigKey(key_name)

    cache_key_name = self._cache.get_key(self._get_helper,
                                         (key_name, value_is_relative_path), {})
    value = self._cache.get(cache_key_name)
    if value is not None:
      return value

    value = _search_key(self._config_dir, key_name, value_is_relative_path)
    if value is None:
      return default

    self._cache.put(cache_key_name, value)
    return value

  def get(self, key_name='', default=None):
    """Get key value using a key name."""
    return self._get_helper(key_name, default=default)

  def get_absolute_path(self, key_name='', default=None):
    """Get absolute path of key value using a key name."""
    return self._get_helper(
        key_name, default=default, value_is_relative_path=True)


class ProjectConfig(Config):
  """Project Config class helper."""

  def __init__(self):
    super(ProjectConfig, self).__init__(PROJECT_PATH)

  def set_environment(self):
    """Sets environment vars from project config."""
    env_variable_values = self.get('env')
    if not env_variable_values:
      return

    for variable, value in six.iteritems(env_variable_values):
      if variable in os.environ:
        # Don't override existing values.
        continue

      os.environ[variable] = str(value)


class AuthConfig(Config):
  """Authentication config."""

  def __init__(self):
    super(AuthConfig, self).__init__(GAE_AUTH_PATH)


class GAEConfig(Config):
  """GAE config."""

  def __init__(self):
    super(GAEConfig, self).__init__(GAE_CONFIG_PATH)


class MonitoringRegionsConfig(Config):
  """Monitoring regions config."""

  def __init__(self):
    super(MonitoringRegionsConfig, self).__init__(MONITORING_REGIONS_PATH)


class IssueTrackerConfig(Config):
  """Issue tracker config."""

  def __init__(self):
    super(IssueTrackerConfig, self).__init__(ISSUE_TRACKERS_PATH)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""libFuzzer engine interface."""

import os
import re
import tempfile

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import dictionary_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import libfuzzer
from clusterfuzz._internal.bot.fuzzers import strategy_selection
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.libFuzzer import constants
from clusterfuzz._internal.bot.fuzzers.libFuzzer import fuzzer
from clusterfuzz._internal.bot.fuzzers.libFuzzer import stats
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.metrics import profiler
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

ENGINE_ERROR_MESSAGE = 'libFuzzer: engine encountered an error'
DICT_PARSING_FAILED_REGEX = re.compile(
    r'ParseDictionaryFile: error in line (\d+)')
MULTISTEP_MERGE_SUPPORT_TOKEN = b'fuzz target overwrites its const input'


def _is_multistep_merge_supported(target_path):
  """Checks whether a particular binary support multistep merge."""
  # TODO(Dor1s): implementation below a temporary workaround, do not tell any
  # body that we are doing this. The real solution would be to execute a
  # fuzz target with '-help=1' and check the output for the presence of
  # multistep merge support added in https://reviews.llvm.org/D71423.
  # The temporary implementation checks that the version of libFuzzer is at
  # least https://github.com/llvm/llvm-project/commit/da3cf61, which supports
  # multi step merge: https://github.com/llvm/llvm-project/commit/f054067.
  if os.path.exists(target_path):
    with open(target_path, 'rb') as file_handle:
      return utils.search_bytes_in_file(MULTISTEP_MERGE_SUPPORT_TOKEN,
                                        file_handle)

  return False


class MergeError(engine.Error):
  """Merge error."""


class LibFuzzerOptions(engine.FuzzOptions):
  """LibFuzzer engine options."""

  def __init__(self, corpus_dir, arguments, strategies, fuzz_corpus_dirs,
               extra_env, use_dataflow_tracing, is_mutations_run):
    super().__init__(corpus_dir, arguments, strategies)
    self.fuzz_corpus_dirs = fuzz_corpus_dirs
    self.extra_env = extra_env
    self.use_dataflow_tracing = use_dataflow_tracing
    self.is_mutations_run = is_mutations_run
    self.merge_back_new_testcases = True
    self.analyze_dictionary = True


class LibFuzzerEngine(engine.Engine):
  """LibFuzzer engine implementation."""

  @property
  def name(self):
    return 'libFuzzer'

  def fuzz_additional_processing_timeout(self, options):
    """Return the maximum additional timeout in seconds for additional
    operations in fuzz() (e.g. merging back new items).

    Args:
      options: A FuzzOptions object.

    Returns:
      An int representing the number of seconds required.
    """
    fuzz_timeout = libfuzzer.get_fuzz_timeout(
        options.is_mutations_run, total_timeout=0)
    # get_fuzz_timeout returns a negative value.
    return -fuzz_timeout

  def prepare(self, corpus_dir, target_path, build_dir):
    """Prepare for a fuzzing session, by generating options. Returns a
    FuzzOptions object.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object.
    """
    del build_dir
    arguments = fuzzer.get_arguments(target_path)
    grammar = fuzzer.get_grammar(target_path)

    if self.do_strategies:
      strategy_pool = strategy_selection.generate_weighted_strategy_pool(
          strategy_list=strategy.LIBFUZZER_STRATEGY_LIST,
          use_generator=True,
          engine_name=self.name)
    else:
      strategy_pool = strategy_selection.StrategyPool()

    strategy_info = libfuzzer.pick_strategies(strategy_pool, target_path,
                                              corpus_dir, arguments, grammar)

    arguments.extend(strategy_info.arguments)

    # Check for seed corpus and add it into corpus directory.
    engine_common.unpack_seed_corpus_if_needed(target_path, corpus_dir)

    # Pick a few testcases from our corpus to use as the initial corpus.
    subset_size = engine_common.random_choice(
        engine_common.CORPUS_SUBSET_NUM_TESTCASES)

    if (not strategy_info.use_dataflow_tracing and
        strategy_pool.do_strategy(strategy.CORPUS_SUBSET_STRATEGY) and
        shell.get_directory_file_count(corpus_dir) > subset_size):
      # Copy |subset_size| testcases into 'subset' directory.
      corpus_subset_dir = self._create_temp_corpus_dir('subset')
      libfuzzer.copy_from_corpus(corpus_subset_dir, corpus_dir, subset_size)
      strategy_info.fuzzing_strategies.append(
          strategy.CORPUS_SUBSET_STRATEGY.name + '_' + str(subset_size))
      strategy_info.additional_corpus_dirs.append(corpus_subset_dir)
    else:
      strategy_info.additional_corpus_dirs.append(corpus_dir)

    # Check dict argument to make sure that it's valid.
    dict_path = fuzzer_utils.extract_argument(
        arguments, constants.DICT_FLAG, remove=False)
    if dict_path and not os.path.exists(dict_path):
      logs.log_error('Invalid dict %s for %s.' % (dict_path, target_path))
      fuzzer_utils.extract_argument(arguments, constants.DICT_FLAG)

    # If there's no dict argument, check for %target_binary_name%.dict file.
    dict_path = fuzzer_utils.extract_argument(
        arguments, constants.DICT_FLAG, remove=False)
    if not dict_path:
      dict_path = dictionary_manager.get_default_dictionary_path(target_path)
      if os.path.exists(dict_path):
        arguments.append(constants.DICT_FLAG + dict_path)

    # If we have a dictionary, correct any items that are not formatted properly
    # (e.g. quote items that are missing them).
    dictionary_manager.correct_if_needed(dict_path)

    strategies = stats.process_strategies(
        strategy_info.fuzzing_strategies, name_modifier=lambda x: x)
    return LibFuzzerOptions(
        corpus_dir, arguments, strategies, strategy_info.additional_corpus_dirs,
        strategy_info.extra_env, strategy_info.use_dataflow_tracing,
        strategy_info.is_mutations_run)

  def _create_empty_testcase_file(self, reproducers_dir):
    """Create an empty testcase file in temporary directory."""
    _, path = tempfile.mkstemp(dir=reproducers_dir)
    return path

  def _create_temp_corpus_dir(self, name):
    """Create temporary corpus directory."""
    new_corpus_directory = os.path.join(fuzzer_utils.get_temp_dir(), name)
    engine_common.recreate_directory(new_corpus_directory)
    return new_corpus_directory

  def _create_merge_corpus_dir(self):
    """Create merge corpus directory."""
    return self._create_temp_corpus_dir('merge-corpus')

  def _merge_new_units(self, target_path, corpus_dir, new_corpus_dir,
                       fuzz_corpus_dirs, arguments, stat_overrides):
    """Merge new units."""
    # Make a decision on whether merge step is needed at all. If there are no
    # new units added by libFuzzer run, then no need to do merge at all.
    new_units_added = shell.get_directory_file_count(new_corpus_dir)
    if not new_units_added:
      stat_overrides['new_units_added'] = 0
      logs.log('Skipped corpus merge since no new units added by fuzzing.')
      return

    # If this times out, it's possible that we will miss some units. However, if
    # we're taking >10 minutes to load/merge the corpus something is going very
    # wrong and we probably don't want to make things worse by adding units
    # anyway.
    merge_corpus = self._create_merge_corpus_dir()

    merge_dirs = fuzz_corpus_dirs[:]

    # Merge the new units with the initial corpus.
    if corpus_dir not in merge_dirs:
      merge_dirs.append(corpus_dir)

    old_corpus_len = shell.get_directory_file_count(corpus_dir)

    new_units_added = 0
    try:
      result = self._minimize_corpus_two_step(
          target_path=target_path,
          arguments=arguments,
          existing_corpus_dirs=merge_dirs,
          new_corpus_dir=new_corpus_dir,
          output_corpus_dir=merge_corpus,
          reproducers_dir=None,
          max_time=engine_common.get_merge_timeout(
              libfuzzer.DEFAULT_MERGE_TIMEOUT))

      libfuzzer.move_mergeable_units(merge_corpus, corpus_dir)
      new_corpus_len = shell.get_directory_file_count(corpus_dir)
      new_units_added = new_corpus_len - old_corpus_len

      stat_overrides.update(result.stats)
    except (MergeError, TimeoutError) as e:
      logs.log_warn('Merge failed.', error=repr(e))

    stat_overrides['new_units_added'] = new_units_added

    # Record the stats to make them easily searchable in stackdriver.
    logs.log('Stats calculated.', stats=stat_overrides)
    if new_units_added:
      logs.log('New units added to corpus: %d.' % new_units_added)
    else:
      logs.log('No new units found.')

  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.

    Returns:
      A FuzzResult object.
    """
    profiler.start_if_needed('libfuzzer_fuzz')
    runner = libfuzzer.get_runner(target_path)
    libfuzzer.set_sanitizer_options(target_path, fuzz_options=options)

    # Directory to place new units.
    if options.merge_back_new_testcases:
      new_corpus_dir = self._create_temp_corpus_dir('new')
      corpus_directories = [new_corpus_dir] + options.fuzz_corpus_dirs
    else:
      corpus_directories = options.fuzz_corpus_dirs

    fuzz_result = runner.fuzz(
        corpus_directories,
        fuzz_timeout=max_time,
        additional_args=options.arguments,
        artifact_prefix=reproducers_dir,
        extra_env=options.extra_env)

    project_qualified_fuzzer_name = (
        engine_common.get_project_qualified_fuzzer_name(target_path))
    dict_error_match = DICT_PARSING_FAILED_REGEX.search(fuzz_result.output)
    if dict_error_match:
      logs.log_error(
          'Dictionary parsing failed (target={target}, line={line}).'.format(
              target=project_qualified_fuzzer_name,
              line=dict_error_match.group(1)),
          engine_output=fuzz_result.output)
    elif (not environment.get_value('USE_MINIJAIL') and
          fuzz_result.return_code == constants.LIBFUZZER_ERROR_EXITCODE):
      # Minijail returns 1 if the exit code is nonzero.
      # Otherwise: we can assume that a return code of 1 means that libFuzzer
      # itself ran into an error.
      logs.log_error(
          ENGINE_ERROR_MESSAGE +
          ' (target={target}).'.format(target=project_qualified_fuzzer_name),
          engine_output=fuzz_result.output)

    log_lines = fuzz_result.output.splitlines()
    # Output can be large, so save some memory by removing reference to the
    # original output which is no longer needed.
    fuzz_result.output = None

    # Check if we crashed, and get the crash testcase path.
    crash_testcase_file_path = runner.get_testcase_path(log_lines)

    # If we exited with a non-zero return code with no crash file in output from
    # libFuzzer, this is most likely a startup crash. Use an empty testcase to
    # to store it as a crash.
    if not crash_testcase_file_path and fuzz_result.return_code:
      crash_testcase_file_path = self._create_empty_testcase_file(
          reproducers_dir)

    # Parse stats information based on libFuzzer output.
    parsed_stats = libfuzzer.parse_log_stats(log_lines)

    # Extend parsed stats by additional performance features.
    parsed_stats.update(
        stats.parse_performance_features(log_lines, options.strategies,
                                         options.arguments))

    # Set some initial stat overrides.
    timeout_limit = fuzzer_utils.extract_argument(
        options.arguments, constants.TIMEOUT_FLAG, remove=False)

    expected_duration = runner.get_max_total_time(max_time)
    actual_duration = int(fuzz_result.time_executed)
    fuzzing_time_percent = 100 * actual_duration / float(expected_duration)
    parsed_stats.update({
        'timeout_limit': int(timeout_limit),
        'expected_duration': expected_duration,
        'actual_duration': actual_duration,
        'fuzzing_time_percent': fuzzing_time_percent,
    })

    # Remove fuzzing arguments before merge and dictionary analysis step.
    non_fuzz_arguments = options.arguments.copy()
    libfuzzer.remove_fuzzing_arguments(non_fuzz_arguments, is_merge=True)

    if options.merge_back_new_testcases:
      self._merge_new_units(target_path, options.corpus_dir, new_corpus_dir,
                            options.fuzz_corpus_dirs, non_fuzz_arguments,
                            parsed_stats)

    fuzz_logs = '\n'.join(log_lines)
    crashes = []
    if crash_testcase_file_path:
      reproduce_arguments = options.arguments[:]
      libfuzzer.remove_fuzzing_arguments(reproduce_arguments)

      # Use higher timeout for reproduction.
      libfuzzer.fix_timeout_argument_for_reproduction(reproduce_arguments)

      # Write the new testcase.
      # Copy crash testcase contents into the main testcase path.
      crashes.append(
          engine.Crash(crash_testcase_file_path, fuzz_logs, reproduce_arguments,
                       actual_duration))

    if options.analyze_dictionary:
      libfuzzer.analyze_and_update_recommended_dictionary(
          runner, project_qualified_fuzzer_name, log_lines, options.corpus_dir,
          non_fuzz_arguments)

    return engine.FuzzResult(fuzz_logs, fuzz_result.command, crashes,
                             parsed_stats, fuzz_result.time_executed)

  def reproduce(self, target_path, input_path, arguments, max_time):
    """Reproduce a crash given an input.

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the reproduction exceeds max_time.
    """
    runner = libfuzzer.get_runner(target_path)
    libfuzzer.set_sanitizer_options(target_path)

    # Remove fuzzing specific arguments. This is only really needed for legacy
    # testcases, and can be removed in the distant future.
    arguments = arguments[:]
    libfuzzer.remove_fuzzing_arguments(arguments)

    runs_argument = constants.RUNS_FLAG + str(constants.RUNS_TO_REPRODUCE)
    arguments.append(runs_argument)

    result = runner.run_single_testcase(
        input_path, timeout=max_time, additional_args=arguments)

    if result.timed_out:
      raise TimeoutError('Reproducing timed out\n' + result.output)

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def _minimize_corpus_two_step(self, target_path, arguments,
                                existing_corpus_dirs, new_corpus_dir,
                                output_corpus_dir, reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      existing_corpus_dirs: Input corpora that existed before the fuzzing run.
      new_corpus_dir: Input corpus that was generated during the fuzzing run.
          Must have at least one new file.
      output_corpus_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A Result object.
    """
    if not _is_multistep_merge_supported(target_path):
      # Fallback to the old single step merge. It does not support incremental
      # stats and provides only `edge_coverage` and `feature_coverage` stats.
      logs.log('Old version of libFuzzer is used. Using single step merge.')
      return self.minimize_corpus(target_path, arguments,
                                  existing_corpus_dirs + [new_corpus_dir],
                                  output_corpus_dir, reproducers_dir, max_time)

    # The dir where merge control file is located must persist for both merge
    # steps. The second step re-uses the MCF produced during the first step.
    merge_control_file_dir = self._create_temp_corpus_dir('mcf_tmp_dir')
    self._merge_control_file = os.path.join(merge_control_file_dir, 'MCF')

    # Two step merge process to obtain accurate stats for the new corpus units.
    # See https://reviews.llvm.org/D66107 for a more detailed description.
    merge_stats = {}

    # Step 1. Use only existing corpus and collect "initial" stats.
    result_1 = self.minimize_corpus(target_path, arguments,
                                    existing_corpus_dirs, output_corpus_dir,
                                    reproducers_dir, max_time)
    merge_stats['initial_edge_coverage'] = result_1.stats['edge_coverage']
    merge_stats['initial_feature_coverage'] = result_1.stats['feature_coverage']

    # Clear the output dir as it does not have any new units at this point.
    engine_common.recreate_directory(output_corpus_dir)

    # Adjust the time limit for the time we spent on the first merge step.
    max_time -= result_1.time_executed
    if max_time <= 0:
      raise TimeoutError('Merging new testcases timed out\n' + result_1.logs)

    # Step 2. Process the new corpus units as well.
    result_2 = self.minimize_corpus(
        target_path, arguments, existing_corpus_dirs + [new_corpus_dir],
        output_corpus_dir, reproducers_dir, max_time)
    merge_stats['edge_coverage'] = result_2.stats['edge_coverage']
    merge_stats['feature_coverage'] = result_2.stats['feature_coverage']

    # Diff the stats to obtain accurate values for the new corpus units.
    merge_stats['new_edges'] = (
        merge_stats['edge_coverage'] - merge_stats['initial_edge_coverage'])
    merge_stats['new_features'] = (
        merge_stats['feature_coverage'] -
        merge_stats['initial_feature_coverage'])

    output = result_1.logs + '\n\n' + result_2.logs
    if (merge_stats['new_edges'] < 0 or merge_stats['new_features'] < 0):
      logs.log_error(
          'Two step merge failed.', merge_stats=merge_stats, output=output)
      merge_stats['new_edges'] = 0
      merge_stats['new_features'] = 0

    self._merge_control_file = None

    # TODO(ochang): Get crashes found during merge.
    return engine.FuzzResult(output, result_2.command, [], merge_stats,
                             result_1.time_executed + result_2.time_executed)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A Result object.

    Raises:
      TimeoutError: If the corpus minimization exceeds max_time.
      Error: If the merge failed in some other way.
    """
    runner = libfuzzer.get_runner(target_path)
    libfuzzer.set_sanitizer_options(target_path)
    merge_tmp_dir = self._create_temp_corpus_dir('merge-workdir')

    result = runner.merge(
        [output_dir] + input_dirs,
        merge_timeout=max_time,
        tmp_dir=merge_tmp_dir,
        additional_args=arguments,
        artifact_prefix=reproducers_dir,
        merge_control_file=getattr(self, '_merge_control_file', None))

    if result.timed_out:
      raise TimeoutError('Merging new testcases timed out\n' + result.output)

    if result.return_code != 0:
      raise MergeError('Merging new testcases failed: ' + result.output)

    merge_output = result.output
    merge_stats = stats.parse_stats_from_merge_log(merge_output.splitlines())

    # TODO(ochang): Get crashes found during merge.
    return engine.FuzzResult(merge_output, result.command, [], merge_stats,
                             result.time_executed)

  def minimize_testcase(self, target_path, arguments, input_path, output_path,
                        max_time):
    """Optional (but recommended): Minimize a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase minimization.
      input_path: Path to the reproducer input.
      output_path: Path to the minimized output.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the testcase minimization exceeds max_time.
    """
    runner = libfuzzer.get_runner(target_path)
    libfuzzer.set_sanitizer_options(target_path)

    minimize_tmp_dir = self._create_temp_corpus_dir('minimize-workdir')
    result = runner.minimize_crash(
        input_path,
        output_path,
        max_time,
        artifact_prefix=minimize_tmp_dir,
        additional_args=arguments)

    if result.timed_out:
      raise TimeoutError('Minimization timed out\n' + result.output)

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def cleanse(self, target_path, arguments, input_path, output_path, max_time):
    """Optional (but recommended): Cleanse a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase cleanse.
      input_path: Path to the reproducer input.
      output_path: Path to the cleansed output.
      max_time: Maximum allowed time for the cleanse.

    Returns:
      A ReproduceResult.

    Raises:
      TimeoutError: If the cleanse exceeds max_time.
    """
    runner = libfuzzer.get_runner(target_path)
    libfuzzer.set_sanitizer_options(target_path)

    cleanse_tmp_dir = self._create_temp_corpus_dir('cleanse-workdir')
    result = runner.cleanse_crash(
        input_path,
        output_path,
        max_time,
        artifact_prefix=cleanse_tmp_dir,
        additional_args=arguments)

    if result.timed_out:
      raise TimeoutError('Cleanse timed out\n' + result.output)

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing module."""

import argparse
import os
import shlex

import clusterfuzz.fuzz


def main():
  os.environ['CF_INTERACTIVE'] = 'True'
  os.environ['PROJECT_NAME'] = 'libClusterFuzz'

  parser = argparse.ArgumentParser(description='Fuzzing tool')
  parser.add_argument('-t', '--target', help='Path to target.', required=True)
  parser.add_argument(
      '-e',
      '--engine',
      help='Fuzzing engine.',
      choices=clusterfuzz.fuzz.ENGINES,
      default='libFuzzer')
  parser.add_argument(
      '-s',
      '--sanitizer',
      help='Sanitizer.',
      choices=['address', 'memory', 'undefined'],
      default='address')
  parser.add_argument('-c', '--corpus', help='Path to corpus.', required=True)
  parser.add_argument(
      '-o', '--output', help='Path to crashers.', default=os.getcwd())
  parser.add_argument(
      '-d',
      '--max-duration',
      help='Max time in seconds to run.',
      type=int,
      default=3600)
  args = parser.parse_args()

  # TODO(ochang): Find a cleaner way to propagate this.
  if args.sanitizer == 'address':
    os.environ['JOB_NAME'] = 'libfuzzer_asan'
  elif args.sanitizer == 'memory':
    os.environ['JOB_NAME'] = 'libfuzzer_msan'
  elif args.sanitizer == 'undefined':
    os.environ['JOB_NAME'] = 'libfuzzer_ubsan'

  os.environ['BUILD_DIR'] = os.path.dirname(args.target)

  engine_impl = clusterfuzz.fuzz.get_engine(args.engine)
  options = engine_impl.prepare(args.corpus, args.target,
                                os.path.dirname(args.target))
  if args.engine == 'libFuzzer':
    options.merge_back_new_testcases = False
    options.analyze_dictionary = False

  result = engine_impl.fuzz(args.target, options, args.output,
                            args.max_duration)
  print('Command: ', ' '.join([shlex.quote(part) for part in result.command]))


if __name__ == '__main__':
  main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for stats."""

import os
import unittest

from clusterfuzz._internal.bot.fuzzers import libfuzzer
from clusterfuzz._internal.bot.fuzzers.libFuzzer import stats
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class PerformanceStatsTest(unittest.TestCase):
  """Performance stats tests class."""

  def setUp(self):
    """Prepare test data and necessary env variables."""
    self.maxDiff = None  # pylint: disable=invalid-name

    test_helpers.patch_environ(self)
    self.data_directory = os.path.join(
        os.path.dirname(__file__), 'libfuzzer_test_data')

  def _read_test_data(self, name):
    """Read test data."""
    data_path = os.path.join(self.data_directory, name)
    with open(data_path) as f:
      return f.read().splitlines()

  def test_parse_stats_from_merge_log(self):
    """Test parsing of a log file produced by libFuzzer run with -merge=1."""
    lines = self._read_test_data('merge_step_1.txt')
    actual_stats = stats.parse_stats_from_merge_log(lines)

    expected_stats = {
        'edge_coverage': 410,
        'feature_coverage': 1869,
    }
    self.assertEqual(expected_stats, actual_stats)

  def test_parse_log_stats(self):
    """Test pure stats parsing without applying of stat_overrides."""
    log_lines = self._read_test_data('no_crash.txt')
    parsed_stats = libfuzzer.parse_log_stats(log_lines)
    expected_stats = {
        'average_exec_per_sec': 97,
        'new_units_added': 55,
        'new_units_generated': 55,
        'number_of_executed_units': 258724,
        'peak_rss_mb': 103,
        'slowest_unit_time_sec': 0
    }

    self.assertEqual(parsed_stats, expected_stats)

  def test_parse_log(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('no_crash_with_strategies.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    expected_stats = {
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'crash_count': 0,
        'corpus_size': 0,
        'dict_used': 1,
        'edge_coverage': 0,
        'edges_total': 398408,
        'feature_coverage': 0,
        'initial_edge_coverage': 0,
        'initial_feature_coverage': 0,
        'leak_count': 0,
        'log_lines_from_engine': 65,
        'log_lines_ignored': 8,
        'log_lines_unwanted': 0,
        'manual_dict_size': 0,
        'max_len': 741802,
        'merge_edge_coverage': 0,
        'new_edges': 0,
        'new_features': 0,
        'oom_count': 0,
        'recommended_dict_size': 0,
        'slow_unit_count': 0,
        'slow_units_count': 0,
        'startup_crash_count': 0,
        'strategy_dataflow_tracing': 0,
        'strategy_corpus_mutations_radamsa': 1,
        'strategy_corpus_mutations_ml_rnn': 0,
        'strategy_corpus_subset': 50,
        'strategy_fork': 1,
        'strategy_mutator_plugin_radamsa': 0,
        'strategy_peach_grammar_mutation': '',
        'strategy_mutator_plugin': 1,
        'strategy_random_max_len': 1,
        'strategy_recommended_dict': 0,
        'strategy_selection_method': 'default',
        'strategy_value_profile': 0,
        'timeout_count': 0
    }

    self.assertDictEqual(expected_stats, parsed_stats)

  def test_parse_log_and_stats_crash(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('crash.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [],
                                                    ['-max_len=1337'])
    self.assertEqual(1, parsed_stats['crash_count'])
    self.assertEqual(0, parsed_stats['corpus_crash_count'])
    self.assertEqual(0, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_go_fuzz(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides for a Go fuzz target."""
    log_lines = self._read_test_data('go_fuzz_log.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [],
                                                    ['-max_len=1337'])
    self.assertEqual(0, parsed_stats['crash_count'])
    self.assertEqual(0, parsed_stats['corpus_crash_count'])
    self.assertEqual(0, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_go_fork_fuzz(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides for a Go fuzz target in fork mode."""
    log_lines = self._read_test_data('go_fork_fuzz_log.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [],
                                                    ['-max_len=1337'])
    self.assertEqual(0, parsed_stats['crash_count'])
    self.assertEqual(0, parsed_stats['corpus_crash_count'])
    self.assertEqual(0, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_fork_fuzz(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides for a fuzz target in fork mode."""
    log_lines = self._read_test_data('fork_fuzz_log.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [],
                                                    ['-max_len=1337'])
    self.assertEqual(0, parsed_stats['crash_count'])
    self.assertEqual(0, parsed_stats['corpus_crash_count'])
    self.assertEqual(0, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_startup_crash(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('startup_crash.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [],
                                                    ['-max_len=1337'])
    self.assertEqual(0, parsed_stats['crash_count'])
    self.assertEqual(0, parsed_stats['corpus_crash_count'])
    self.assertEqual(1, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_corpus_crash(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('corpus_crash.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    self.assertEqual(1, parsed_stats['crash_count'])
    self.assertEqual(1, parsed_stats['corpus_crash_count'])
    self.assertEqual(0, parsed_stats['startup_crash_count'])

  def test_parse_log_and_stats_corpus_crash_with_corpus_subset(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('corpus_crash_with_corpus_subset.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    self.assertEqual(1, parsed_stats['strategy_corpus_subset'])

  def test_parse_log_and_stats_oom(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('oom.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    self.assertEqual(1, parsed_stats['oom_count'])
    self.assertEqual(0, parsed_stats['timeout_count'])

  def test_parse_log_and_stats_oom_in_seed_corpus(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('oom_in_seed_corpus.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    self.assertEqual(1, parsed_stats['oom_count'])
    self.assertEqual(0, parsed_stats['timeout_count'])

  def test_parse_log_and_stats_from_corrupted_output(self):
    """Test stats parsing from a log with corrupted libFuzzer stats."""
    log_lines = self._read_test_data('corrupted_stats.txt')
    parsed_stats = libfuzzer.parse_log_stats(log_lines)
    self.assertNotIn('peak_rss_mb', parsed_stats)

  def test_parse_log_and_stats_timeout(self):
    """Test stats parsing and additional performance features extraction
    without applying of stat_overrides."""
    log_lines = self._read_test_data('timeout.txt')
    parsed_stats = stats.parse_performance_features(log_lines, [], [])
    self.assertEqual(0, parsed_stats['oom_count'])
    self.assertEqual(1, parsed_stats['timeout_count'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for monitor."""
# pylint: disable=protected-access

import os
import unittest

from clusterfuzz._internal.metrics import monitor
from clusterfuzz._internal.tests.test_libs import helpers


class InitializeTest(unittest.TestCase):
  """Tests for initialize."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.config.local_config.ProjectConfig.get',
        'clusterfuzz._internal.metrics.monitor.check_module_loaded',
        'google.cloud.monitoring_v3.MetricServiceClient',
        'threading.Thread.start',
    ])
    helpers.patch_environ(self, {})
    self.mock.check_module_loaded.return_value = True
    self.mock.get.return_value = True

    os.environ['BOT_NAME'] = 'somebot'

  def test_local(self):
    """Tests no initialization on local development."""
    os.environ['LOCAL_DEVELOPMENT'] = '1'
    monitor.initialize()
    self.assertEqual(0, self.mock.start.call_count)

  def test_disabled(self):
    """Tests no initialization when monitoring is disabled in config."""
    self.mock.get.return_value = False
    monitor.initialize()
    self.assertEqual(0, self.mock.start.call_count)

  def test_error(self):
    """Tests error."""
    self.mock.check_module_loaded.return_value = False
    monitor.initialize()
    self.assertEqual(0, self.mock.start.call_count)

  def test_initialize(self):
    """Tests initialization."""
    monitor.initialize()
    self.assertEqual(1, self.mock.start.call_count)


class MonitorTest(unittest.TestCase):
  """Tests for monitor functions."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.monitor.check_module_loaded',
    ])
    self.mock.check_module_loaded.return_value = True
    monitor.metrics_store().reset_for_testing()

  def test_counter_metric_success(self):
    self.assertIsInstance(
        monitor.CounterMetric('t', 'desc', field_spec=None),
        monitor._CounterMetric)

  def test_counter_metric_failure(self):
    """Test counter metric failure."""
    self.mock.check_module_loaded.return_value = False
    counter = monitor.CounterMetric(
        't', 'desc', field_spec=[
            monitor.StringField('name'),
        ])
    counter.increment()
    counter.increment({'name': 'test'})
    counter.increment_by(5)
    self.assertIsInstance(counter, monitor._MockMetric)

  def test_cumulative_distribution_metric_success(self):
    self.assertIsInstance(
        monitor.CumulativeDistributionMetric(
            'd', 'desc', bucketer=None, field_spec=None),
        monitor._CumulativeDistributionMetric)

  def test_cumulative_distribution_metric_failure(self):
    self.mock.check_module_loaded.return_value = False
    counter = monitor.CumulativeDistributionMetric(
        'd', 'desc', bucketer=None, field_spec=None)
    counter.add(10)
    self.assertIsInstance(counter, monitor._MockMetric)

  def test_get_region(self):
    """Ensure it assigns the right region"""
    self.assertEqual('linux', monitor._get_region('clusterfuzz-linux-0123'))
    self.assertEqual('linux', monitor._get_region('clusterfuzz-linux-hijk'))
    self.assertEqual('linux-high-end',
                     monitor._get_region('clusterfuzz-linux-high-end-0017'))
    self.assertEqual('linux-pre',
                     monitor._get_region('clusterfuzz-linux-pre-0123'))
    self.assertEqual('windows', monitor._get_region('clusterfuzz-windows-0010'))
    self.assertEqual('windows-pre',
                     monitor._get_region('clusterfuzz-windows-pre-0017'))
    self.assertEqual('android', monitor._get_region('clusterfuzz-android-1337'))
    self.assertEqual('mac', monitor._get_region('clusterfuzz-mac-1337'))
    self.assertEqual('unknown', monitor._get_region('diwejfwlejf'))

  def test_cumulative_distribution_metric_fixed(self):
    """Test _CumulativeDistributionMetric with fixed bucketer."""
    # Buckets:
    #   [-Inf, 0.1)
    #   [0.1, 0.2)
    #   [0.2, 0.3)
    #   ...
    #   [1.0, Inf)
    metric = monitor.CumulativeDistributionMetric(
        'name',
        description='test metric',
        bucketer=monitor.FixedWidthBucketer(width=0.1, num_finite_buckets=10),
        field_spec=None)

    metric.add(-1)
    metric.add(0.05)
    metric.add(0.1)
    metric.add(0.15)
    metric.add(0.3)
    metric.add(0.75)
    metric.add(1)
    metric.add(2)

    result = monitor.metrics_store().get(metric, None).value

    self.assertListEqual([
        1,
        1,
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        2,
    ], result.buckets)

  def test_cumulative_distribution_metric_geometric(self):
    """Test _CumulativeDistributionMetric with geometric bucketer."""
    # Buckets:
    #   [-Inf, 1)
    #   [1, 2)
    #   [2, 4)
    #   [4, 8)
    #   [8, 16)
    #   [16, 32)
    #   [32, 64)
    #   [64, Inf)
    metric = monitor.CumulativeDistributionMetric(
        'name',
        description='test metric',
        bucketer=monitor.GeometricBucketer(
            scale=1.0, growth_factor=2, num_finite_buckets=6),
        field_spec=None)

    metric.add(0)
    metric.add(1)
    metric.add(2)
    metric.add(3)
    metric.add(12)
    metric.add(17)
    metric.add(31)
    metric.add(32.1)
    metric.add(40)
    metric.add(64.1)
    result = monitor.metrics_store().get(metric, None).value

    self.assertListEqual([
        1,
        1,
        2,
        0,
        1,
        2,
        2,
        1,
    ], result.buckets)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for ML training tasks."""

import os

from clusterfuzz._internal.fuzzing import corpus_manager
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell


def get_corpus(corpus_directory, fuzzer_name):
  """Get corpus directory.

  This function will download latest corpus backup file from GCS, unzip
  the file and put them in corpus directory.

  Args:
    directory: The directory to place corpus.
    fuzzer_name: Fuzzer name, e.g. libpng_read_fuzzer, xml_parser_fuzzer, etc.

  Returns:
    True if the corpus can be acquired and False otherwise.
  """
  backup_bucket_name = environment.get_value('BACKUP_BUCKET')
  corpus_fuzzer_name = environment.get_value('CORPUS_FUZZER_NAME_OVERRIDE')

  # Get GCS backup path.
  gcs_backup_path = corpus_manager.gcs_url_for_backup_file(
      backup_bucket_name, corpus_fuzzer_name, fuzzer_name,
      corpus_manager.LATEST_BACKUP_TIMESTAMP)

  # Get local backup path.
  local_backup_name = os.path.basename(gcs_backup_path)
  local_backup_path = os.path.join(corpus_directory, local_backup_name)

  # Download latest backup.
  if not storage.copy_file_from(gcs_backup_path, local_backup_path):
    logs.log_error(
        'Failed to download corpus from GCS bucket {}.'.format(gcs_backup_path))
    return False

  # Extract corpus from zip file.
  archive.unpack(local_backup_path, corpus_directory)
  shell.remove_file(local_backup_path)
  return True


def get_gcs_model_directory(folder, fuzzer_name):
  """
  Get gcs bucket path to store latest model.

  Args:
    folder (str): Subdirectory denoting the category of ML model being trained.
    fuzzer_name (str): Exactly what gets passed in to `execute_task()`.

  Returns:
    A string with the GCS absolute path to upload the model to.
  """
  model_bucket_name = environment.get_value('CORPUS_BUCKET')
  if not model_bucket_name:
    return None

  gcs_model_directory = 'gs://{}/{}/{}'.format(model_bucket_name, folder,
                                               fuzzer_name)

  return gcs_model_directory



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for managing and interacting with Fuchsia devices
via undercoat."""

import os
import shutil
import tempfile

from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process

# In order to properly clean up any stale instances, we keep track of handles in
# the persistent cache. We assume that only one instance of a bot will be
# accessing this cache at a time.
HANDLE_CACHE_KEY = 'undercoat-handles'


def add_running_handle(handle):
  """Record a handle as potentially needing to be cleaned up on restart."""
  new_handle_list = list(set(get_running_handles()) | set([handle]))
  persistent_cache.set_value(
      HANDLE_CACHE_KEY, new_handle_list, persist_across_reboots=True)


def remove_running_handle(handle):
  """Remove a handle from the tracked set."""
  new_handle_list = list(set(get_running_handles()) - set([handle]))
  persistent_cache.set_value(
      HANDLE_CACHE_KEY, new_handle_list, persist_across_reboots=True)


def get_running_handles():
  """Get a list of potentially stale handles from previous runs."""
  return persistent_cache.get_value(HANDLE_CACHE_KEY, default_value=[])


def get_temp_dir():
  """Define a tempdir for undercoat to store its data in.

  This tempdir needs to be of a scope that persists across invocations of the
  bot, to ensure proper cleanup of stale handles/data."""
  temp_dir = os.path.join(environment.get_value('ROOT_DIR'), 'bot', 'undercoat')
  os.makedirs(temp_dir, exist_ok=True)

  return temp_dir


class UndercoatError(Exception):
  """Error for errors while running undercoat."""


def undercoat_api_command(*args):
  """Make an API call to the undercoat binary."""
  logs.log(f'Running undercoat command {args}')
  bundle_dir = environment.get_value('FUCHSIA_RESOURCES_DIR')
  undercoat_path = os.path.join(bundle_dir, 'undercoat', 'undercoat')
  undercoat = new_process.ProcessRunner(undercoat_path, args)
  # The undercoat log is sent to stderr, which we capture to a tempfile
  with tempfile.TemporaryFile() as undercoat_log:
    result = undercoat.run_and_wait(
        stderr=undercoat_log, extra_env={'TMPDIR': get_temp_dir()})
    result.output = result.output.decode('utf-8')

    if result.return_code != 0:
      # Dump the undercoat log to assist in debugging
      log_data = utils.read_from_handle_truncated(undercoat_log, 1024 * 1024)
      logs.log_warn('Log output from undercoat: ' + log_data.decode('utf-8'))

      # The API error message is returned on stdout
      raise UndercoatError(
          'Error running undercoat command %s: %s' % (args, result.output))

  return result


def undercoat_instance_command(command, handle, *args, abort_on_error=True):
  """Helper for the subset of undercoat commands that operate on an instance."""
  try:
    return undercoat_api_command(command, '-handle', handle, *args)
  except UndercoatError:
    if abort_on_error:
      # Try to print extra logs and shut down
      # TODO(eep): Should we be attempting to automatically restart?
      dump_instance_logs(handle)
      stop_instance(handle)
    raise


def get_version():
  """Get undercoat API version as (major, minor, patch) tuple."""
  version = undercoat_api_command('version').output

  if not version.startswith('v'):
    raise UndercoatError('Invalid version reported: %s' % version)

  parts = version[1:].split('.')
  if len(parts) != 3:
    raise UndercoatError('Invalid version reported: %s' % version)

  try:
    return tuple(int(part) for part in parts)
  except ValueError as e:
    raise UndercoatError('Invalid version reported: %s' % version) from e


def validate_api_version():
  """Check that the undercoat API version is supported. Raises an error if it is
  not."""
  major, minor, patch = get_version()
  if major > 0:
    raise UndercoatError(
        'Unsupported API version: %d.%d.%d' % (major, minor, patch))


def dump_instance_logs(handle):
  """Dump logs from an undercoat instance."""
  qemu_log = undercoat_instance_command(
      'get_logs', handle, abort_on_error=False).output
  logs.log_warn(qemu_log)


def start_instance():
  """Start an instance via undercoat."""
  handle = undercoat_api_command('start_instance').output.strip()
  logs.log('Started undercoat instance with handle %s' % handle)

  # Immediately save the handle in case we crash before stop_instance()
  # is called
  add_running_handle(handle)

  return handle


def stop_all():
  """Attempt to stop any running undercoat instances that may have not been
  cleanly shut down."""
  for handle in get_running_handles():
    try:
      undercoat_instance_command('stop_instance', handle, abort_on_error=False)
    except UndercoatError:
      pass

    # Even if we failed to stop_instance above, there's no point in trying
    # again later
    remove_running_handle(handle)

  # At this point, all handles/data should have been cleaned up, but if any is
  # remaining then we clear it out here
  shutil.rmtree(get_temp_dir())


def stop_instance(handle):
  """Stop a running undercoat instance."""
  result = undercoat_instance_command(
      'stop_instance', handle, abort_on_error=False)

  # Mark the corresponding handle as having been cleanly shut down
  remove_running_handle(handle)

  return result


def list_fuzzers(handle):
  """Start an instance via undercoat."""
  return undercoat_instance_command('list_fuzzers', handle).output.split('\n')


def prepare_fuzzer(handle, fuzzer):
  """Prepare a fuzzer of the given name for use, via undercoat."""
  return undercoat_instance_command('prepare_fuzzer', handle, '-fuzzer', fuzzer)


def run_fuzzer(handle, fuzzer, outdir, args):
  """Run a fuzzer of the given name, via undercoat."""
  # TODO(fxbug.dev/47490): Pass back raw return code from libFuzzer?
  undercoat_args = ['-fuzzer', fuzzer]
  if outdir:
    undercoat_args += ['-artifact-dir', outdir]
  args = undercoat_args + ['--'] + args
  return undercoat_instance_command('run_fuzzer', handle, *args)


def put_data(handle, fuzzer, src, dst):
  """Put files for a fuzzer onto an instance, via undercoat.
  If src is a directory, it will be copied recursively. Standard globs are
  supported."""
  return undercoat_instance_command('put_data', handle, '-fuzzer', fuzzer,
                                    '-src', src, '-dst', dst).output


def get_data(handle, fuzzer, src, dst):
  """Get files from a fuzzer on an instance, via undercoat.
  If src is a directory, it will be copied recursively. Standard globs are
  supported."""
  try:
    return undercoat_instance_command(
        'get_data',
        handle,
        '-fuzzer',
        fuzzer,
        '-src',
        src,
        '-dst',
        dst,
        abort_on_error=False).output
  except UndercoatError:
    return None
